# A couple tests to make sure that versioning works as expected within Travis
# So, for example, the build would fail before accidentally publishing a bad version
from unittest.mock import patch


# Mocking out getenv() instead of actually setting envars so this doesn't affect other tests
@patch("pyinaturalist.getenv", side_effect=["true", "master", "123"])
def test_version__stable_release(mock_getenv):
    import pyinaturalist

    assert "dev" not in pyinaturalist.__version__


@patch("pyinaturalist.getenv", side_effect=["true", "dev", "123"])
def test_version__pre_release(mock_getenv):
    import pyinaturalist

    assert pyinaturalist.get_prerelease_version("1.0.0") == "1.0.0-dev.123"



"""
Utilities built on top of ``python-forge`` used to simplify usage for API docs, mainly by
combining function signature modification with docstring modification.
"""
from inspect import cleandoc
from itertools import chain
from functools import wraps
from logging import getLogger
from typing import Callable, List

from pyinaturalist.constants import TemplateFunction

logger = getLogger(__name__)


def document_request_params(template_functions: List[TemplateFunction]):
    """Document a function with both docstrings and function signatures from one or more
    template functions.

    Signature modification requires ``python-forge``. If not installed, only docstrings will be modified.

    Example:

        >>> # 1. Template function with individual request params + docs
        >>> def get_foo_template(arg_1: str = None, arg_2: bool = False):
        >>>     '''
        >>>     arg_1: Example request parameter 1
        >>>     arg_2: Example request parameter 2
        >>>     '''

        >>> # 2. Decorated API Endpoint function with generic (variadic) keyword args
        >>> @document_request_params(get_foo_template)
        >>> def get_foo(**kwargs) -> List:
            ''' Get Foo resource '''

        >>> # 3. Modified function signature + docstring
        >>> help(get_foo)
        '''
        Help on function get_foo:
        get_foo(arg_1: str = None, arg_2: bool = False) -> List
        Get Foo resource
        Args:
            arg_1: Example request parameter 1
            arg_2: Example request parameter 2
        '''

    Args:
        template_functions: Template functions containing docstrings and params to apply to the
            wrapped function
    """
    template_functions = [*template_functions, _user_agent]

    def f(func):
        # Modify docstring
        func = copy_docstrings(func, template_functions)

        # If forge is installed, modify signature; otherwise, silently ignore it
        try:
            func = copy_signatures(func, template_functions)
        except ImportError:
            logger.debug("Forge not installed; skipping runtime API documentation")

        # Call modified function
        @wraps(func)
        def g(*args, **kwargs):
            return func(*args, **kwargs)

        return g

    return f


def copy_docstrings(
    target_function: Callable, template_functions: List[TemplateFunction]
) -> Callable:
    """Copy docstrings from one or more template functions to a target function.
    Assumes Google-style docstrings.

    Args:
        target_function: Function to modify
        template_functions: Functions containing docstrings to apply to ``target_function``
    """
    # Start with initial function description only
    docstring, return_value_desc = _split_docstring(target_function.__doc__ or "")

    # Insert 'Args' section
    docstring += "\n\nArgs:"
    for template_function in template_functions:
        docstring += template_function.__doc__ or ""
        docstring = docstring.rstrip()

    # Insert 'Returns' section, if present
    if return_value_desc:
        docstring += "\n\nReturns:\n    " + return_value_desc

    target_function.__doc__ = docstring
    return target_function


def _split_docstring(docstring):
    """ Split a docstring into a function description + return value description, if present. """
    if "Returns:" in docstring:
        function_desc, return_value_desc = docstring.split("Returns:")
    else:
        function_desc = docstring
        return_value_desc = ""

    return cleandoc(function_desc.strip()), cleandoc(return_value_desc.strip())


def copy_signatures(
    target_function: Callable, template_functions: List[TemplateFunction]
) -> Callable:
    """Copy function signatures from one or more template functions to a target function.

    Args:
        target_function: Function to modify
        template_functions: Functions containing params to apply to ``target_function``
    """
    revision = _get_combined_revision(template_functions)
    return revision(target_function)


def _get_combined_revision(template_functions: List[TemplateFunction]):
    """ Create a :py:class:`forge.Revision` from the combined parameters of multiple functions """
    import forge

    # Use forge.copy to create a revision for each template function
    revisions = [forge.copy(func) for func in template_functions]

    # Combine the parameters of all revisions into a single revision
    fparams = [list(rev.signature.parameters.values()) for rev in revisions]
    return forge.sign(*list(chain.from_iterable(fparams)))


# Param template that's added to every function signature by default
def _user_agent(user_agent: str = None):
    """
    user_agent: A custom user-agent string to provide to the iNaturalist API
    """



""" Helper functions for formatting API responses """
from typing import Any, Dict, List, Iterable, Optional


def as_geojson_feature_collection(
    results: Iterable[Dict[str, Any]], properties: List[str] = None
) -> Dict[str, Any]:
    """
    Convert results from an API response into a
    `geojson FeatureCollection <https://tools.ietf.org/html/rfc7946#section-3.3>`_ object.
    This is currently only used for observations, but could be used for any other responses with
    geospatial info.

    Args:
        results: List of results from API response
        properties: Whitelist of specific properties to include
    """
    return {
        "type": "FeatureCollection",
        "features": [as_geojson_feature(record, properties) for record in results],
    }


def as_geojson_feature(result: Dict[str, Any], properties: List[str] = None) -> Dict[str, Any]:
    """
    Convert an individual response item to a geojson Feature object, optionally with specific
    response properties included.

    Args:
        result: A single response item
        properties: Whitelist of specific properties to include
    """
    result["geojson"]["coordinates"] = [float(i) for i in result["geojson"]["coordinates"]]
    return {
        "type": "Feature",
        "geometry": result["geojson"],
        "properties": {k: result.get(k) for k in properties or []},
    }


def flatten_nested_params(observation: Dict[str, Any]) -> Dict[str, Any]:
    """Extract some nested observation properties to include at the top level;
     this makes it easier to specify these as properties for
     :py:func:`.as_as_geojson_feature_collection`.

    Args:
        observation: A single observation result
    """
    taxon = observation.get("taxon", {})
    photos = observation.get("photos", [{}])
    observation["taxon_id"] = taxon.get("id")
    observation["taxon_name"] = taxon.get("name")
    observation["taxon_rank"] = taxon.get("rank")
    observation["preferred_common_name"] = taxon.get("preferred_common_name")
    observation["photo_url"] = photos[0].get("url")
    return observation


def format_taxon(taxon: Dict) -> str:
    """Format a taxon result into a single string containing taxon ID, rank, and name
    (including common name, if available).
    """
    # Visually align taxon IDs (< 7 chars) and ranks (< 11 chars)
    common = taxon.get("preferred_common_name")
    return "{:>8}: {:>12} {}{}".format(
        taxon["id"],
        taxon["rank"].title(),
        taxon["name"],
        " ({})".format(common) if common else "",
    )


def convert_location_to_float(results: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Convert coordinate pairs in response items from strings to floats, if valid.

    Args:
        results: Results from API response; expects coordinates in the "location" key
    """
    for result in results or []:
        if "," in (result["location"] or ""):
            result["location"] = [convert_float(coord) for coord in result["location"].split(",")]
    return results


def convert_lat_long_to_float(results: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Convert coordinate pairs in response items from strings to floats, if valid

    Args:
        results: Results from API response; expects coordinates in "latitude" and "longitude" keys
    """
    for result in results:
        result["latitude"] = convert_float(result["latitude"])
        result["longitude"] = convert_float(result["longitude"])
    return results


def convert_float(value: Any) -> Optional[float]:
    """ Convert a value to a float, if valid; return ``None`` otherwise """
    try:
        return float(value)
    except (TypeError, ValueError):
        return None



"""
Reusable template functions used for API documentation.
Each template function contains a portion of an endpoint's request parameters, with corresponding
type annotations and docstrings.
"""
from typing import Any, Dict, List, Iterable

from pyinaturalist.constants import (
    MultiInt,
    MultiStr,
    Date,
    DateTime,
    IntOrStr,
    FileOrPath,
    ObsFieldValues,
)
from pyinaturalist.request_params import MULTIPLE_CHOICE_PARAMS


# Observations
# --------------------


# Params that are in most observation-related endpoints in both Node and REST APIs
def _observation_common(
    q: str = None,
    d1: Date = None,
    d2: Date = None,
    day: MultiInt = None,
    month: MultiInt = None,
    year: MultiInt = None,
    license: MultiStr = None,
    list_id: int = None,
    photo_license: MultiStr = None,
    out_of_range: bool = None,
    quality_grade: str = None,
    taxon_id: MultiInt = None,
    taxon_name: MultiStr = None,
    iconic_taxa: MultiStr = None,
    updated_since: DateTime = None,
):
    """
    q: Search observation properties
    d1: Must be observed on or after this date
    d2: Must be observed on or before this date
    day: Must be observed within this day of the month
    month: Must be observed within this month
    year: Must be observed within this year
    license: Observation must have this license
    photo_license: Must have at least one photo with this license
    out_of_range: Observations whose taxa are outside their known ranges
    list_id: Taxon must be in the list with this ID
    quality_grade: Must have this quality grade
    taxon_id: Only show observations of these taxa and their descendants
    taxon_name: Taxon must have a scientific or common name matching this string
    iconic_taxa: Taxon must by within this iconic taxon
    updated_since: Must be updated since this time
    """


# Observation params that are only in the Node API
def _observation_node_only(
    acc: bool = None,
    captive: bool = None,
    endemic: bool = None,
    geo: bool = None,
    id_please: bool = None,
    identified: bool = None,
    introduced: bool = None,
    mappable: bool = None,
    native: bool = None,
    pcid: bool = None,
    photos: bool = None,
    popular: bool = None,
    sounds: bool = None,
    taxon_is_active: bool = None,
    threatened: bool = None,
    verifiable: bool = None,
    id: MultiInt = None,
    not_id: MultiInt = None,
    sound_license: MultiStr = None,
    ofv_datatype: MultiStr = None,
    place_id: MultiInt = None,
    project_id: MultiInt = None,
    rank: MultiStr = None,
    site_id: MultiStr = None,
    without_taxon_id: MultiInt = None,
    user_id: MultiInt = None,
    user_login: MultiStr = None,
    term_id: MultiInt = None,
    term_value_id: MultiInt = None,
    without_term_value_id: MultiInt = None,
    acc_above: str = None,
    acc_below: str = None,
    created_d1: DateTime = None,
    created_d2: DateTime = None,
    created_on: Date = None,
    observed_on: Date = None,
    unobserved_by_user_id: int = None,
    apply_project_rules_for: str = None,
    cs: str = None,
    csa: str = None,
    csi: MultiStr = None,
    geoprivacy: MultiStr = None,
    taxon_geoprivacy: MultiStr = None,
    max_rank: str = None,
    min_rank: str = None,
    hrank: str = None,
    lrank: str = None,
    id_above: int = None,
    id_below: int = None,
    identifications: str = None,
    lat: float = None,
    lng: float = None,
    radius: float = None,
    not_in_project: IntOrStr = None,
    not_matching_project_rules_for: IntOrStr = None,
    search_on: str = None,
    viewer_id: int = None,
    reviewed: bool = None,
    locale: str = None,
    preferred_place_id: int = None,
    ttl: str = None,
):
    """
    acc: Whether or not positional accuracy / coordinate uncertainty has been specified
    captive: Captive or cultivated observations
    endemic: Observations whose taxa are endemic to their location
    geo: Observations that are georeferenced
    id_please: Observations with the **deprecated** "ID, Please!" flag.
        Note that this will return observations, but that this attribute is no longer used.
    identified: Observations that have community identifications
    introduced: Observations whose taxa are introduced in their location
    mappable: Observations that show on map tiles
    native: Observations whose taxa are native to their location
    pcid: Observations identified by the curator of a project. If the ``project_id`` parameter
        is also specified, this will only consider observations identified by curators of the
        specified project(s)
    photos: Observations with photos
    popular: Observations that have been favorited by at least one user
    sounds: Observations with sounds
    taxon_is_active: Observations of active taxon concepts
    threatened: Observations whose taxa are threatened in their location
    verifiable: Observations with a ``quality_grade`` of either ``needs_id`` or ``research``.
        Equivalent to ``quality_grade=needs_id,research``
    id: Must have this ID
    not_id: Must not have this ID
    place_id: Must be observed within the place with this ID
    project_id: Must be added to the project this ID or slug
    rank: Taxon must have this rank
    site_id: Must be affiliated with the iNaturalist network website with this ID
    ofv_datatype: Must have an observation field value with this datatype
    sound_license: Must have at least one sound with this license
    without_taxon_id: Exclude observations of these taxa and their descendants
    user_id: User must have this ID or login
    user_login: User must have this login
    term_id: Must have an annotation using this controlled term ID
    term_value_id: Must have an annotation using this controlled value ID.
        Must be combined with the ``term_id`` parameter
    without_term_value_id: Exclude observations with annotations using this controlled value ID.
        Must be combined with the ``term_id`` parameter
    acc_above: Must have an positional accuracy above this value (meters)
    acc_below: Must have an positional accuracy below this value (meters)
    created_d1: Must be created at or after this time
    created_d2: Must be created at or before this time
    created_on: Must be created on this date
    observed_on: Must be observed on this date
    unobserved_by_user_id: Must not be of a taxon previously observed by this user
    apply_project_rules_for: Must match the rules of the project with this ID or slug
    cs: Taxon must have this conservation status code. If the ``place_id`` parameter is also
        specified, this will only consider statuses specific to that place
    csa: Taxon must have a conservation status from this authority. If the ``place_id`` parameter is
        also specified, this will only consider statuses specific to that place
    csi: Taxon must have this IUCN conservation status. If the ``place_id`` parameter is also
        specified, this will only consider statuses specific to that place
    geoprivacy: Must have this geoprivacy setting
    taxon_geoprivacy: Filter observations by the most conservative geoprivacy applied by a
        conservation status associated with one of the taxa proposed in the current
        identifications.
    hrank: Taxon must have this rank or lower
    lrank: Taxon must have this rank or higher
    id_above: Must have an ID above this value
    id_below: Must have an ID below this value
    identifications: Identifications must meet these criteria
    lat: Must be within a ``radius`` kilometer circle around this lat/lng (lat, lng, radius)
    lng: Must be within a ``radius`` kilometer circle around this lat/lng (lat, lng, radius)
    radius: Must be within a {radius} kilometer circle around this lat/lng (lat, lng, radius)
    not_in_project: Must not be in the project with this ID or slug
    not_matching_project_rules_for: Must not match the rules of the project with this ID or slug
    search_on: Properties to search on, when combined with q. Searches across all properties by
        default
    viewer_id: See reviewed
    reviewed: Observations have been reviewed by the user with ID equal to the value of the
        ``viewer_id`` parameter
    locale: Locale preference for taxon common names
    preferred_place_id: Place preference for regional taxon common names
    ttl: Set the ``Cache-Control`` HTTP header with this value as ``max-age``, in seconds
    """


# Observation params that are only in the REST API
def _observation_rest_only(
    has: MultiStr = None,
    on: Date = None,
    m1: Date = None,
    m2: Date = None,
    h1: Date = None,
    h2: Date = None,
    extra: str = None,
    response_format: str = "json",
):
    """
    has: Catch-all for some boolean selectors. This can be used multiple times, e.g.
        ``has=['photos', 'geo']``
    m1: First month of a month range
    m2: Last month of a month range
    h1: First hour of an hour range
    h2: Last hour of an hour range
    on: Filter by date string
    extra: Retrieve additional information.
        **'projects'** returns info about the projects the observations have been added to,
        **'fields'** returns observation field values,
        **'observation_photos'** returns information about the photos' relationship with the
        observation, like their order.
    response_format: A supported response format to return
    """


def _create_observation(
    species_guess: str = None,
    taxon_id: int = None,
    observed_on_string: Date = None,
    time_zone: str = None,
    description: str = None,
    tag_list: MultiStr = None,
    place_guess: str = None,
    latitude: float = None,
    longitude: float = None,
    map_scale: int = None,
    positional_accuracy: int = None,
    geoprivacy: str = None,
    observation_fields: ObsFieldValues = None,
    observation_field_values_attributes: ObsFieldValues = None,
    flickr_photos: MultiInt = None,
    picasa_photos: MultiStr = None,
    facebook_photos: MultiStr = None,
    local_photos: Iterable[FileOrPath] = None,
):
    """
    species_guess: Equivalent to the "What did you see?" field on the observation form.
        iNat will try to choose a single taxon based on this, but it may fail if it's ambuguous
    taxon_id: ID of the taxon to associate with this observation
    observed_on_string: Date/time of the observation. Time zone will default to the user's
        time zone if not specified.
    time_zone: Time zone the observation was made in
    description: Observation description
    tag_list: Comma-separated list of tags
    place_guess: Name of the place where the observation was recorded.
        **Note:** iNat will *not* try to automatically look up coordinates based on this string
    latitude: Latitude of the observation; presumed datum is **WGS84**
    longitude: Longitude of the observation; presumed datum is **WGS84**
    map_scale: Google Maps zoom level (from **0 to 19**) at which to show this observation's map marker.
    positional_accuracy: Positional accuracy of the observation coordinates, in meters
    geoprivacy: Geoprivacy for the observation
    observation_fields: Dict of observation fields in the format ``{id: value}``.
        Alias for ``observation_field_values_attributes``.
    flickr_photos: Flickr photo ID(s) to add as photos for this observation. User must have
        their Flickr and iNat accounts connected, and the user must own the photo(s) on Flickr.
    picasa_photos: Picasa photo ID(s) to add as photos for this observation. User must have
        their Picasa and iNat accounts connected, and the user must own the photo(s) on Picasa.
    facebook_photos: Facebook photo IDs to add as photos for this observation. User must have
        their Facebook and iNat accounts connected, and the user must own the photo on Facebook.
    local_photos: Image files, file-like objects, and/or paths for local photos to upload
    """


def _update_observation(
    # _method: str = None,  # Exposed as a client-specific workaround; not needed w/ `requests`
    ignore_photos: bool = True,
):
    """
    ignore_photos
        If photos exist on the observation but are missing in the request, simply ignore them
        instead of deleting the missing observation photos
    """


# Projects
# --------------------


# Note: 'page' parameter is not in API docs, but it appears to work. 'order' does not, however.
def _projects_params(
    q: str = None,
    id: MultiInt = None,
    not_id: MultiInt = None,
    lat: float = None,
    lng: float = None,
    radius: int = 500,
    featured: bool = None,
    noteworthy: bool = None,
    site_id: int = None,
    rule_details: bool = None,
    type: MultiStr = None,
    member_id: int = None,
    has_params: bool = None,
    has_posts: bool = None,
):
    """
    q: Project name must begin with this value
    id: Must have this ID
    not_id: Must not have this ID
    lat: Must be within a ``radius`` kilometer circle around this lat/lng
    lng: Must be within a ``radius`` kilometer circle around this lat/lng
    radius: Distance from center (``(lat, lng)``) to search, in kilometers. Defaults to 500km.
    featured:  Must be marked featured for the relevant site
    noteworthy: Must be marked noteworthy for the relevant site
    site_id: Site ID that applies to ``featured`` and ``noteworthy``.
        Defaults to the site of the authenticated user, or to the main iNaturalist site
    rule_details: Return more information about project rules, for example return a full taxon
        object instead of simply an ID
    type: Projects must be of this type
    member_id: Project must have member with this user ID
    has_params: Must have search parameter requirements
    has_posts: Must have posts
    per_page: Number of results to return in a page. The maximum value is generally 200 unless otherwise noted
    order_by: Sort order.
        ``distance`` only applies if lat and lng are specified.
        ``featured`` only applies if ``featured`` or ``noteworthy`` are true.
    """


# Taxa
# --------------------


def _taxon_params(
    q: str = None,
    is_active: bool = None,
    taxon_id: int = None,
    rank: str = None,
    max_rank: str = None,
    min_rank: str = None,
    rank_level: int = None,
    locale: str = None,
    preferred_place_id: int = None,
    all_names: bool = None,
    per_page: int = None,
):
    """
    q: Name must begin with this value
    is_active: Taxon is active
    taxon_id: Only show taxa with this ID, or its descendants
    rank: Taxon must have this exact rank
    min_rank: Taxon must have this rank or higher; overrides ``rank``
    max_rank: Taxon must have this rank or lower; overrides ``rank``
    rank_level: Taxon must have this rank level. Some example values are 70 (kingdom), 60 (phylum),
        50 (class), 40 (order), 30 (family), 20 (genus), 10 (species), 5 (subspecies)
    locale: Locale preference for taxon common names
    preferred_place_id: Place preference for regional taxon common names
    all_names: Include all taxon names in the response
    per_page: Number of results to return in a page. The maximum value is generally 200 unless
        otherwise noted
    """


def _taxon_id_params(
    id_above: int = None,
    id_below: int = None,
    only_id: int = None,
    parent_id: int = None,
):
    """
    id_above: Must have an ID above this value
    id_below: Must have an ID below this value
    only_id: Return only the record IDs
    parent_id: Taxon's parent must have this ID
    """


# Individual/common params
# ------------------------


def _access_token(access_token: str = None):
    """
    access_token: An access token required for user authentication, as returned by
        :py:func:`get_access_token()`
    """


def _bounding_box(
    nelat: float = None,
    nelng: float = None,
    swlat: float = None,
    swlng: float = None,
):
    """
    nelat: NE latitude of bounding box
    nelng: NE longitude of bounding box
    swlat: SW latitude of bounding box
    swlng: SW longitude of bounding box
    """


def _geojson_properties(properties: List[str] = None):
    """
    properties: Properties from observation results to include as GeoJSON properties
    """


# TODO: Remove deprecated `params` arg in 0.12
def _legacy_params(params: Dict[str, Any] = None):
    """
    params: [DEPRECATED] Request parameters as a dict instead of keyword arguments
    """


def _minify(minify: str = None):
    """
    minify: Condense each match into a single string containing taxon ID, rank, and name
    """


def _name(name: str = None):
    """
    name: Name must match this value
    """


def _only_id(only_id: bool = False):
    """
    only_id: Return only the record IDs
    """


def _observation_id(observation_id: int):
    """
    observation_id: iNaturalist observation ID
    """


def _page(page: int = None):
    """
    page: Page number of results to return
    """


def _pagination(
    page: int = None,
    per_page: int = None,
    order: str = None,
    order_by: str = None,
):
    """
    page: Page number of results to return
    per_page: Number of results to return in a page. The maximum value is generally 200,
        unless otherwise noted
    order: Sort order
    order_by: Field to sort on
    """


_get_observations = [
    _legacy_params,
    _observation_common,
    _observation_node_only,
    _bounding_box,
]


# TODO: Remove deprecated `search_query` kwarg in 0.12
def _search_query(q: str = None, search_query: str = None):
    """
    q: Search query
    """


def _format_param_choices():
    return "\n".join(
        ["  * {}: {}".format(param, choices) for param, choices in MULTIPLE_CHOICE_PARAMS.items()]
    )


MULTIPLE_CHOICE_PARAM_DOCS = "**Multiple-Choice Parameters:**\n" + _format_param_choices()



"""
Code to access the (read-only, but fast) Node based public iNaturalist API
See: http://api.inaturalist.org/v1/docs/

Most recent API version tested: 1.3.0

Functions
---------

.. automodsumm:: pyinaturalist.node_api
    :functions-only:
    :nosignatures:

"""
from logging import getLogger
from time import sleep
from typing import List

import requests
from urllib.parse import urljoin

from pyinaturalist import api_docs as docs
from pyinaturalist.constants import (
    INAT_NODE_API_BASE_URL,
    PER_PAGE_RESULTS,
    THROTTLING_DELAY,
    MultiInt,
    JsonResponse,
    RequestParams,
)
from pyinaturalist.exceptions import ObservationNotFound
from pyinaturalist.forge_utils import document_request_params
from pyinaturalist.request_params import (
    DEFAULT_OBSERVATION_ATTRS,
    NODE_OBS_ORDER_BY_PROPERTIES,
    PROJECT_ORDER_BY_PROPERTIES,
    check_deprecated_params,
    translate_rank_range,
    validate_multiple_choice_param,
)
from pyinaturalist.response_format import (
    format_taxon,
    as_geojson_feature_collection,
    flatten_nested_params,
    convert_location_to_float,
)
from pyinaturalist.api_requests import get

logger = getLogger(__name__)


def make_inaturalist_api_get_call(endpoint: str, **kwargs) -> requests.Response:
    """Make an API call to iNaturalist.

    Args:
        endpoint: The name of an endpoint not including the base URL e.g. 'observations'
        kwargs: Arguments for :py:func:`.api_requests.request`
    """
    return get(urljoin(INAT_NODE_API_BASE_URL, endpoint), **kwargs)


# Observations
# --------------------


def get_observation(observation_id: int, user_agent: str = None) -> JsonResponse:
    """Get details about an observation.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Observations/get_observations_id

    Example:

        >>> get_observation(16227955)

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation.json
                :language: JSON

    Args:
        observation_id: Observation ID
        user_agent: a user-agent string that will be passed to iNaturalist.

    Returns:
        A dict with details on the observation

    Raises:
        :py:exc:`.ObservationNotFound`
    """

    r = get_observations(id=observation_id, user_agent=user_agent)
    if r["results"]:
        return r["results"][0]

    raise ObservationNotFound()


@document_request_params([*docs._get_observations, docs._pagination, docs._only_id])
def get_observations(
    params: RequestParams = None, user_agent: str = None, **kwargs
) -> JsonResponse:
    """Search observations.

    **API reference:** http://api.inaturalist.org/v1/docs/#!/Observations/get_observations

    Example:

        >>> # Get observations of Monarch butterflies with photos + public location info,
        >>> # on a specific date in the provice of Saskatchewan, CA
        >>> observations = get_observations(
        >>>     taxon_name='Danaus plexippus',
        >>>     created_on='2020-08-27',
        >>>     photos=True,
        >>>     geo=True,
        >>>     geoprivacy='open',
        >>>     place_id=7953,
        >>> )

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observations_node_page1.json
                :language: JSON

    Returns:
        JSON response containing observation records
    """
    kwargs = check_deprecated_params(params, **kwargs)
    validate_multiple_choice_param(kwargs, "order_by", NODE_OBS_ORDER_BY_PROPERTIES)
    r = make_inaturalist_api_get_call("observations", params=kwargs, user_agent=user_agent)
    return r.json()


@document_request_params([*docs._get_observations, docs._only_id])
def get_all_observations(
    params: RequestParams = None, user_agent: str = None, **kwargs
) -> List[JsonResponse]:
    """Like :py:func:`get_observations()`, but handles pagination and returns all results in one
    call. Explicit pagination parameters will be ignored.

    Notes on pagination from the iNaturalist documentation:
    "The large size of the observations index prevents us from supporting the page parameter when
    retrieving records from large result sets. If you need to retrieve large numbers of records,
    use the ``per_page`` and ``id_above`` or ``id_below`` parameters instead."

    Example:

        >>> observations = get_all_observations(
        >>>     taxon_name='Danaus plexippus',
        >>>     created_on='2020-08-27',
        >>> )

    Returns:
        Combined list of observation records. Response format is the same as the inner "results"
        object returned by :py:func:`.get_observations()`.
    """
    kwargs = check_deprecated_params(params, **kwargs)
    results = []  # type: List[JsonResponse]
    id_above = 0
    pagination_params = {
        **kwargs,
        **{
            "order_by": "id",
            "order": "asc",
            "per_page": PER_PAGE_RESULTS,
            "user_agent": user_agent,
        },
    }

    while True:
        pagination_params["id_above"] = id_above
        page_obs = get_observations(**pagination_params)
        results = results + page_obs.get("results", [])

        if page_obs["total_results"] <= PER_PAGE_RESULTS:
            return results

        sleep(THROTTLING_DELAY)
        id_above = results[-1]["id"]


@document_request_params([*docs._get_observations, docs._pagination])
def get_observation_species_counts(user_agent: str = None, **kwargs) -> JsonResponse:
    """Get all species (or other "leaf taxa") associated with observations matching the search
    criteria, and the count of observations they are associated with.
    **Leaf taxa** are the leaves of the taxonomic tree, e.g., species, subspecies, variety, etc.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Observations/get_observations_species_counts

    Example:
        >>> get_observation_species_counts(user_login='my_username', quality_grade='research')

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_species_counts.json
                :language: JSON

    Returns:
        JSON response containing taxon records with counts
    """
    r = make_inaturalist_api_get_call(
        "observations/species_counts",
        params=kwargs,
        user_agent=user_agent,
    )
    r.raise_for_status()
    return r.json()


@document_request_params(docs._get_observations)
def get_all_observation_species_counts(user_agent: str = None, **kwargs) -> List[JsonResponse]:
    """Like :py:func:`get_observation_species_counts()`, but handles pagination and returns all
    results in one call. Explicit pagination parameters will be ignored.

    Notes:
    While the ``page`` parameter is undocumented for observations/species_counts, it appears to be supported.
    ``id_above`` and ``id_below`` are not helpful in the context.

    Example:
        >>> get_all_observation_species_counts(
        ...     user_agent=None,
        ...     quality_grade='research',
        ...     place_id=154695,
        ...     iconic_taxa='Reptilia',
        ... )

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_all_observation_species_counts_ex_results.json
                :language: JSON

    Returns:
        Combined list of taxon records with counts
    """
    kwargs = check_deprecated_params(**kwargs)
    results = []  # type: List[JsonResponse]
    page = 1

    pagination_params = {
        **kwargs,
        **{
            "per_page": PER_PAGE_RESULTS,
            "user_agent": user_agent,
        },
    }

    while True:
        pagination_params["page"] = page
        page_obs = get_observation_species_counts(**pagination_params)
        results = results + page_obs.get("results", [])

        if len(results) == page_obs["total_results"]:
            return results

        sleep(THROTTLING_DELAY)
        page += 1


@document_request_params([*docs._get_observations, docs._geojson_properties])
def get_geojson_observations(properties: List[str] = None, **kwargs) -> JsonResponse:
    """Get all observation results combined into a GeoJSON ``FeatureCollection``.
    By default this includes some basic observation properties as GeoJSON ``Feature`` properties.
    The ``properties`` argument can be used to override these defaults.

    Example:
        >>> get_geojson_observations(observation_id=16227955, properties=["photo_url"])

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observations.geojson
                :language: JSON

    Returns:
        A ``FeatureCollection`` containing observation results as ``Feature`` dicts.
    """
    kwargs["mappable"] = True
    observations = get_all_observations(**kwargs)
    return as_geojson_feature_collection(
        (flatten_nested_params(obs) for obs in observations),
        properties=properties if properties is not None else DEFAULT_OBSERVATION_ATTRS,
    )


# Places
# --------------------

# TODO: Use updated sample responses with converted coords


def _convert_all_locations_to_float(response):
    """ Convert locations for both standard (curated) and community-contributed places to floats """
    response["results"] = {
        "standard": convert_location_to_float(response["results"].get("standard")),
        "community": convert_location_to_float(response["results"].get("community")),
    }
    return response


def get_places_by_id(place_id: MultiInt, user_agent: str = None) -> JsonResponse:
    """
    Get one or more places by ID.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Places/get_places_id

    Example:
        >>> get_places_by_id([93735, 89191])

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_places_by_id.json
                :language: JSON

    Args:
        place_id: Get a place with this ID. Multiple values are allowed.

    Returns:
        JSON response containing place records
    """
    r = make_inaturalist_api_get_call("places", ids=place_id, user_agent=user_agent)
    r.raise_for_status()

    # Convert coordinates to floats
    response = r.json()
    response["results"] = convert_location_to_float(response["results"])
    return response


@document_request_params([docs._bounding_box, docs._name])
def get_places_nearby(user_agent: str = None, **kwargs) -> JsonResponse:
    """
    Given an bounding box, and an optional name query, return standard iNaturalist curator approved
    and community non-curated places nearby

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Places/get_places_nearby

    Example:
        >>> bounding_box = (150.0, -50.0, -149.999, -49.999)
        >>> get_places_nearby(*bounding_box)

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_places_nearby.json
                :language: JSON

    Returns:
        JSON response containing place records
    """
    r = make_inaturalist_api_get_call("places/nearby", params=kwargs, user_agent=user_agent)
    r.raise_for_status()
    return _convert_all_locations_to_float(r.json())


def get_places_autocomplete(q: str, user_agent: str = None) -> JsonResponse:
    """Given a query string, get places with names starting with the search term

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Places/get_places_autocomplete

    Example:
        >>> get_places_autocomplete('Irkutsk')

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_places_autocomplete.json
                :language: JSON

    Args:
        q: Name must begin with this value

    Returns:
        JSON response containing place records
    """
    r = make_inaturalist_api_get_call("places/autocomplete", params={"q": q}, user_agent=user_agent)
    r.raise_for_status()

    # Convert coordinates to floats
    response = r.json()
    response["results"] = convert_location_to_float(response["results"])
    return response


# Projects
# --------------------


@document_request_params([docs._projects_params, docs._pagination])
def get_projects(user_agent: str = None, **kwargs) -> JsonResponse:
    """Given zero to many of following parameters, get projects matching the search criteria.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Projects/get_projects

    Example:

        Search for projects about invasive species within 400km of Vancouver, BC:

        >>> response = get_projects(
        >>>     q='invasive',
        >>>     lat=49.27,
        >>>     lng=-123.08,
        >>>     radius=400,
        >>>     order_by='distance',
        >>> )

        Show basic info for projects in response:

        >>> project_info = {p["id"]: p["title"] for p in response["results"]}
        >>> print('\\n'.join([f'{k}:\\t{v}' for k, v in project_info.items()]))
        8291:    PNW Invasive Plant EDDR
        19200:   King County (WA) Noxious and Invasive Weeds
        8527:    WA Invasives
        2344:    Invasive & Huntable Animals
        6432:    CBWN Invasive Plants

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_projects.json
                :language: JSON

    Returns:
        JSON response containing project records
    """
    validate_multiple_choice_param(kwargs, "order_by", PROJECT_ORDER_BY_PROPERTIES)
    r = make_inaturalist_api_get_call("projects", params=kwargs, user_agent=user_agent)
    r.raise_for_status()

    # Convert coordinates to floats
    response = r.json()
    response["results"] = convert_location_to_float(response["results"])
    return response


def get_projects_by_id(
    project_id: MultiInt, rule_details: bool = None, user_agent: str = None
) -> JsonResponse:
    """Get one or more projects by ID.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Projects/get_projects_id

    Example:

        >>> get_projects_by_id([8348, 6432])

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_projects_by_id.json
                :language: JSON

    Args:
        project_id: Get projects with this ID. Multiple values are allowed.
        rule_details: Return more information about project rules, for example return a full taxon
            object instead of simply an ID

    Returns:
        JSON response containing project records
    """
    r = make_inaturalist_api_get_call(
        "projects",
        ids=project_id,
        params={"rule_details": rule_details},
        user_agent=user_agent,
    )
    r.raise_for_status()

    # Convert coordinates to floats
    response = r.json()
    response["results"] = convert_location_to_float(response["results"])
    return response


# Taxa
# --------------------


@document_request_params([docs._taxon_params, docs._taxon_id_params])
def get_taxa(user_agent: str = None, **kwargs) -> JsonResponse:
    """Given zero to many of following parameters, get taxa matching the search criteria.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Taxa/get_taxa

    Example:

        >>> response = get_taxa(q="vespi", rank=["genus", "family"])
        >>> print({taxon["id"]: taxon["name"] for taxon in response["results"]})
        {52747: "Vespidae", 84737: "Vespina", 92786: "Vespicula", 646195: "Vespiodes", ...}

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_taxa.json
                :language: JSON

    Returns:
        JSON response containing taxon records
    """
    kwargs = translate_rank_range(kwargs)
    r = make_inaturalist_api_get_call("taxa", params=kwargs, user_agent=user_agent)
    r.raise_for_status()
    return r.json()


def get_taxa_by_id(taxon_id: MultiInt, user_agent: str = None) -> JsonResponse:
    """Get one or more taxa by ID.

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Taxa/get_taxa_id

    Example:

        >>> response = get_taxa_by_id(343248)
        >>> basic_fields = ["preferred_common_name", "observations_count", "wikipedia_url", "wikipedia_summary"]
        >>> print({f: response["results"][0][f] for f in basic_fields})
        {
            "preferred_common_name": "Paper Wasps",
            "observations_count": 69728,
            "wikipedia_url": "http://en.wikipedia.org/wiki/Polistinae",
            "wikipedia_summary": "The Polistinae are eusocial wasps closely related to the more familiar yellow jackets...",
        }

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_taxa_by_id.json
                :language: JSON

    Args:
        taxon_id: Get taxa with this ID. Multiple values are allowed.

    Returns:
        JSON response containing taxon records
    """
    r = make_inaturalist_api_get_call("taxa", ids=taxon_id, user_agent=user_agent)
    r.raise_for_status()
    return r.json()


@document_request_params([docs._taxon_params, docs._minify])
def get_taxa_autocomplete(user_agent: str = None, **kwargs) -> JsonResponse:
    """Given a query string, returns taxa with names starting with the search term

    **API reference:** https://api.inaturalist.org/v1/docs/#!/Taxa/get_taxa_autocomplete

    **Note:** There appears to currently be a bug in the API that causes ``per_page`` to not have
    any effect.

    Example:

        >>> response = get_taxa_autocomplete(q='vespi')
        >>> first_result = response['results'][0]
        >>> print(first_result['rank'], first_result['name'])
        'family Vespidae'

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_taxa_autocomplete.json
                :language: JSON

        .. admonition:: Example Response (with **minify=True**)
            :class: toggle

            .. literalinclude:: ../sample_data/get_taxa_autocomplete_minified.json
                :language: JSON

    Returns:
        JSON response containing taxon records
    """
    kwargs = translate_rank_range(kwargs)
    r = make_inaturalist_api_get_call("taxa/autocomplete", params=kwargs, user_agent=user_agent)
    r.raise_for_status()
    json_response = r.json()

    if kwargs.get("minify"):
        json_response["results"] = [format_taxon(t) for t in json_response["results"]]
    return json_response



from logging import getLogger
from os import getenv

__author__ = "Nicolas Noé"
__email__ = "nicolas@niconoe.eu"
__version__ = "0.11.0"

# These are imported here so they can be set with pyinaturalist.<variable>
from pyinaturalist.constants import DRY_RUN_ENABLED, DRY_RUN_WRITE_ONLY

DEFAULT_USER_AGENT = "Pyinaturalist/{version}".format(version=__version__)
user_agent = DEFAULT_USER_AGENT


def get_prerelease_version(version: str) -> str:
    """If we're running in a Travis CI job on the dev branch, get a prerelease version using the
    current build number. For example: ``1.0.0 -> 1.0.0-dev.123``

    This could also be done in ``.travis.yml``, but it's a bit cleaner to do in python.
    """
    if not (getenv("TRAVIS") == "true" and getenv("TRAVIS_BRANCH") == "dev"):
        return version
    new_version = "{}-dev.{}".format(version, getenv("TRAVIS_BUILD_NUMBER", "0"))
    getLogger(__name__).info("Using pre-release version: {}".format(new_version))
    return new_version


# This won't modify the version outside of Travis
__version__ = get_prerelease_version(__version__)



#!/usr/bin/env python
"""A semi-automated script used to test all observation CRUD endpoints.
Must provide iNat credentials via environment variables. Usage example:
```
export INAT_USERNAME=""
export INAT_PASSWORD=""
export INAT_APP_ID=""
export INAT_APP_SECRET=""
python test/manual_tests/obs_crud_test.py
```
"""
from datetime import datetime
from os import getenv
from os.path import join
from pprint import pprint

from pyinaturalist.rest_api import (
    add_photo_to_observation,
    create_observation,
    delete_observation,
    get_access_token,
    put_observation_field_values,
    update_observation,
)
from pyinaturalist.node_api import get_observation
from test.conftest import SAMPLE_DATA_DIR

SAMPLE_PHOTO = join(SAMPLE_DATA_DIR, "obs_image.jpg")


def run_observation_crud_test():
    # TODO: Built-in support for using envars for auth instead of function args might be useful
    token = get_access_token(
        username=getenv("INAT_USERNAME"),
        password=getenv("INAT_PASSWORD"),
        app_id=getenv("INAT_APP_ID"),
        app_secret=getenv("INAT_APP_SECRET"),
    )
    print("Received access token")

    test_obs_id = create_test_obs(token)
    update_test_obs(test_obs_id, token)
    delete_test_obs(test_obs_id, token)
    print("Test complete")


def create_test_obs(token):
    response = create_observation(
        taxon_id=54327,
        observed_on_string=datetime.now().isoformat(),
        description="This is a test observation used by pyinaturalist, and will be deleted shortly.",
        tag_list="wasp, Belgium",
        latitude=50.647143,
        longitude=4.360216,
        positional_accuracy=50,
        geoprivacy="open",
        access_token=token,
        observation_fields={297: 1},
    )
    test_obs_id = response[0]["id"]
    print("Created new observation: {}".format(test_obs_id))

    obs = get_observation(test_obs_id)
    print("Fetched new observation:")
    pprint(obs, indent=2)
    return test_obs_id


def update_test_obs(test_obs_id, token):
    response = add_photo_to_observation(
        test_obs_id,
        photo=SAMPLE_PHOTO,
        access_token=token,
    )
    photo_id = response.get("photo").get("id")
    assert photo_id
    print("Added photo to observation: {}".format(photo_id))
    # pprint(response, indent=2)

    response = update_observation(
        test_obs_id,
        geoprivacy="obscured",
        ignore_photos=True,
        access_token=token,
    )
    new_geoprivacy = response[0]["geoprivacy"]
    assert new_geoprivacy == "obscured"
    print("Updated observation")
    # pprint(response, indent=2)

    response = put_observation_field_values(
        observation_id=test_obs_id,
        observation_field_id=297,
        value=2,
        access_token=token,
    )
    print("Added observation field value:")
    pprint(response, indent=2)


def delete_test_obs(test_obs_id, token):
    response = delete_observation(test_obs_id, token)
    # Empty response is expected
    print("Deleted observation")
    pprint(response, indent=2)


if __name__ == "__main__":
    run_observation_crud_test()



"""
Code used to access the (read/write, but slow) Rails based API of iNaturalist
See: https://www.inaturalist.org/pages/api+reference

Functions
---------

.. automodsumm:: pyinaturalist.rest_api
    :functions-only:
    :nosignatures:

"""
from time import sleep
from typing import Dict, Any, List, Union

from urllib.parse import urljoin

from pyinaturalist import api_docs as docs
from pyinaturalist.constants import (
    THROTTLING_DELAY,
    INAT_BASE_URL,
    FileOrPath,
    JsonResponse,
    ListResponse,
    RequestParams,
)
from pyinaturalist.exceptions import AuthenticationError, ObservationNotFound
from pyinaturalist.api_requests import delete, get, post, put
from pyinaturalist.forge_utils import document_request_params
from pyinaturalist.request_params import (
    OBSERVATION_FORMATS,
    REST_OBS_ORDER_BY_PROPERTIES,
    check_deprecated_params,
    convert_observation_fields,
    ensure_file_obj,
    ensure_file_objs,
    validate_multiple_choice_param,
    warn,
)
from pyinaturalist.response_format import convert_lat_long_to_float


def get_access_token(
    username: str, password: str, app_id: str, app_secret: str, user_agent: str = None
) -> str:
    """Get an access token using the user's iNaturalist username and password.
    You still need an iNaturalist app to do this.

    **API reference:** https://www.inaturalist.org/pages/api+reference#auth

    Example:
        >>> access_token = get_access_token('...')
        >>> headers = {"Authorization": f"Bearer {access_token}"}

    Args:
        username: iNaturalist username
        password: iNaturalist password
        app_id: iNaturalist application ID
        app_secret: iNaturalist application secret
        user_agent: a user-agent string that will be passed to iNaturalist.
    """
    payload = {
        "client_id": app_id,
        "client_secret": app_secret,
        "grant_type": "password",
        "username": username,
        "password": password,
    }

    response = post(
        "{base_url}/oauth/token".format(base_url=INAT_BASE_URL),
        json=payload,
        user_agent=user_agent,
    )
    try:
        return response.json()["access_token"]
    except KeyError:
        raise AuthenticationError("Authentication error, please check credentials.")


@document_request_params(
    [
        docs._observation_common,
        docs._observation_rest_only,
        docs._bounding_box,
        docs._pagination,
    ]
)
def get_observations(user_agent: str = None, **kwargs) -> Union[List, str]:
    """Get observation data, optionally in an alternative format. Also see
    :py:func:`.get_geojson_observations` for GeoJSON format (not included here because it wraps
    a separate API endpoint).

    **API reference:** https://www.inaturalist.org/pages/api+reference#get-observations

    Example:

        >>> get_observations(id=45414404, format="atom")

        .. admonition:: Example Response (atom)
            :class: toggle

            .. literalinclude:: ../sample_data/get_observations.atom
                :language: xml

        .. admonition:: Example Response (csv)
            :class: toggle

            .. literalinclude:: ../sample_data/get_observations.csv

        .. admonition:: Example Response (dwc)
            :class: toggle

            .. literalinclude:: ../sample_data/get_observations.dwc
                :language: xml

        .. admonition:: Example Response (json)
            :class: toggle

            .. literalinclude:: ../sample_data/get_observations.json
                :language: json

        .. admonition:: Example Response (kml)
            :class: toggle

            .. literalinclude:: ../sample_data/get_observations.kml
                :language: xml

        .. admonition:: Example Response (widget)
            :class: toggle

            .. literalinclude:: ../sample_data/get_observations.js
                :language: javascript

    Returns:
        Return type will be ``dict`` for the ``json`` response format, and ``str`` for all
        others.
    """
    response_format = kwargs.pop("response_format", "json")
    if response_format == "geojson":
        raise ValueError("For geojson format, use pyinaturalist.node_api.get_geojson_observations")
    if response_format not in OBSERVATION_FORMATS:
        raise ValueError("Invalid response format")
    validate_multiple_choice_param(kwargs, "order_by", REST_OBS_ORDER_BY_PROPERTIES)

    response = get(
        urljoin(INAT_BASE_URL, "observations.{}".format(response_format)),
        params=kwargs,
        user_agent=user_agent,
    )

    if response_format == "json":
        return convert_lat_long_to_float(response.json())
    else:
        return response.text


@document_request_params([docs._search_query, docs._page])
def get_observation_fields(user_agent: str = None, **kwargs) -> ListResponse:
    """Search observation fields. Observation fields are basically typed data fields that
    users can attach to observation.

    **API reference:** https://www.inaturalist.org/pages/api+reference#get-observation_fields

    Example:

        >>> get_observation_fields(q='number of individuals')
        >>> # Show just observation field IDs and names
        >>> from pprint import pprint
        >>> pprint({r['id']: r['name'] for r in response})

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/get_observation_fields_page1.json
                :language: javascript

    Returns:
        Observation fields as a list of dicts
    """
    kwargs = check_deprecated_params(**kwargs)
    response = get(
        "{base_url}/observation_fields.json".format(base_url=INAT_BASE_URL),
        params=kwargs,
        user_agent=user_agent,
    )
    return response.json()


@document_request_params([docs._search_query])
def get_all_observation_fields(**kwargs) -> ListResponse:
    """
    Like :py:func:`.get_observation_fields()`, but handles pagination for you.

    Example:

        >>> get_all_observation_fields(q='number of')

    Returns:
        Observation fields as a list of dicts. Response format is the same as the inner
        "results" object returned by :py:func:`.get_observation_fields`.
    """
    results = []  # type: List[Dict[str, Any]]
    page = 1

    while True:
        r = get_observation_fields(page=page, **kwargs)

        if not r:
            return results

        results += r
        page += 1
        sleep(THROTTLING_DELAY)


def put_observation_field_values(
    observation_id: int,
    observation_field_id: int,
    value: Any,
    access_token: str,
    user_agent: str = None,
) -> JsonResponse:
    # TODO: Also implement a put_or_update_observation_field_values() that deletes then recreates the field_value?
    # TODO: Return some meaningful exception if it fails because the field is already set.
    # TODO: It appears pushing the same value/pair twice in a row (but deleting it meanwhile via the UI)...
    # TODO: ...triggers an error 404 the second time (report to iNaturalist?)
    """Set an observation field (value) on an observation.
    Will fail if this observation field is already set for this observation.

    To find an `observation_field_id`, either user :py:func:`.get_observation_fields` or search
    on iNaturalist: https://www.inaturalist.org/observation_fields

    **API reference:** https://www.inaturalist.org/pages/api+reference#put-observation_field_values-id

    Example:
            >>> # First find an observation field by name, if the ID is unknown
            >>> response = get_observation_fields('vespawatch_id')
            >>> observation_field_id = response[0]['id']
            >>>
            >>> put_observation_field_values(
            >>>     observation_id=7345179,
            >>>     observation_field_id=observation_field_id,
            >>>     value=250,
            >>>     access_token=token,
            >>> )

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/put_observation_field_value_result.json
                :language: javascript

    Args:
        observation_id: ID of the observation receiving this observation field value
        observation_field_id: ID of the observation field for this observation field value
        value: Value for the observation field
        access_token: access_token: The access token, as returned by :func:`get_access_token()`
        user_agent: A user-agent string that will be passed to iNaturalist.

    Returns:
        The nwely updated field value record
    """

    payload = {
        "observation_field_value": {
            "observation_id": observation_id,
            "observation_field_id": observation_field_id,
            "value": value,
        }
    }

    response = put(
        "{base_url}/observation_field_values/{id}".format(
            base_url=INAT_BASE_URL, id=observation_field_id
        ),
        access_token=access_token,
        user_agent=user_agent,
        json=payload,
    )

    response.raise_for_status()
    return response.json()


def create_observations(params: RequestParams = None, **kwargs):
    """Create a new observation.
    Note: Creating multiple observations sould be possible according to the docs, but it does not
    appear to work.
    """
    warn(
        "create_observations() has been deprecated, as creating multiple observations is not "
        "currently functional. Please use create_observation() instead."
    )
    create_observation(params, **kwargs)


# TODO: more thorough usage example
@document_request_params([docs._legacy_params, docs._access_token, docs._create_observation])
def create_observation(
    params: RequestParams = None, access_token: str = None, user_agent: str = None, **kwargs
) -> ListResponse:
    """Create a new observation.

    **API reference:** https://www.inaturalist.org/pages/api+reference#post-observations

    Example:
        >>> token = get_access_token('...')
        >>> create_observation(
        >>>     access_token=token,
        >>>     species_guess='Pieris rapae',
        >>>     local_photos='~/observation_photos/2020_09_01_14003156.jpg',
        >>>     observation_fields={297: 1},  # 297 is the obs. field ID for 'Number of individuals'
        >>> )

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/create_observation_result.json
                :language: javascript

        .. admonition:: Example Response (failure)
            :class: toggle

            .. literalinclude:: ../sample_data/create_observation_fail.json
                :language: javascript

    Returns:
        JSON response containing the newly created observation(s)

    Raises:
        :py:exc:`requests.HTTPError`, if the call is not successful. iNaturalist returns an
        error 422 (unprocessable entity) if it rejects the observation data (for example an
        observation date in the future or a latitude > 90. In that case the exception's
        ``response`` attribute gives more details about the errors.
    """
    # Accept either top-level params (like most other endpoints)
    # or nested {"observation": params} (like the iNat API accepts directly)
    if "observation" in kwargs:
        kwargs.update(kwargs.pop("observation"))
    kwargs = check_deprecated_params(params, **kwargs)
    kwargs = convert_observation_fields(kwargs)
    if "local_photos" in kwargs:
        kwargs["local_photos"] = ensure_file_objs(kwargs["local_photos"])

    response = post(
        url="{base_url}/observations.json".format(base_url=INAT_BASE_URL),
        json={"observation": kwargs},
        access_token=access_token,
        user_agent=user_agent,
    )
    response.raise_for_status()
    return response.json()


@document_request_params(
    [
        docs._observation_id,
        docs._legacy_params,
        docs._access_token,
        docs._create_observation,
        docs._update_observation,
    ]
)
def update_observation(
    observation_id: int,
    params: RequestParams = None,
    access_token: str = None,
    user_agent: str = None,
    **kwargs
) -> ListResponse:
    """
    Update a single observation.

    **API reference:** https://www.inaturalist.org/pages/api+reference#put-observations-id

    .. note::

        Unlike the underlying REST API endpoint, this function will **not** delete any existing
        photos from your observation if not specified in ``local_photos``. If you want this to
        behave the same as the REST API and you do want to delete photos, call with
        ``ignore_photos=False``.

    Example:

        >>> token = get_access_token('...')
        >>> update_observation(
        >>>     17932425,
        >>>     access_token=token,
        >>>     description="updated description!",
        >>> )

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/update_observation_result.json
                :language: javascript

    Returns:
        JSON response containing the newly updated observation(s)

    Raises:
        :py:exc:`requests.HTTPError`, if the call is not successful. iNaturalist returns an
            error 410 if the observation doesn't exists or belongs to another user.
    """
    # Accept either top-level params (like most other endpoints)
    # or nested params (like the iNat API actually accepts)
    if "observation" in kwargs:
        kwargs.update(kwargs.pop("observation"))
    kwargs = check_deprecated_params(params, **kwargs)
    kwargs = convert_observation_fields(kwargs)
    if "local_photos" in kwargs:
        kwargs["local_photos"] = ensure_file_objs(kwargs["local_photos"])

    # This is the one Boolean parameter that's specified as an int, for some reason.
    # Also, set it to True if not specified, which seems like much saner default behavior.
    if "ignore_photos" in kwargs:
        kwargs["ignore_photos"] = int(kwargs["ignore_photos"])
    else:
        kwargs["ignore_photos"] = 1

    response = put(
        url="{base_url}/observations/{id}.json".format(base_url=INAT_BASE_URL, id=observation_id),
        json={"observation": kwargs},
        access_token=access_token,
        user_agent=user_agent,
    )
    response.raise_for_status()
    return response.json()


def add_photo_to_observation(
    observation_id: int,
    photo: FileOrPath,
    access_token: str,
    user_agent: str = None,
):
    """Upload a local photo and assign it to an existing observation.

    **API reference:** https://www.inaturalist.org/pages/api+reference#post-observation_photos

    Example:

        >>> token = get_access_token('...')
        >>> add_photo_to_observation(
        >>>     1234,
        >>>     '~/observation_photos/2020_09_01_14003156.jpg',
        >>>     access_token=token,
        >>> )

        .. admonition:: Example Response
            :class: toggle

            .. literalinclude:: ../sample_data/add_photo_to_observation.json
                :language: javascript

    Args:
        observation_id: the ID of the observation
        photo: An image file, file-like object, or path
        access_token: the access token, as returned by :func:`get_access_token()`
        user_agent: a user-agent string that will be passed to iNaturalist.

    Returns:
        Information about the newly created photo
    """
    response = post(
        url="{base_url}/observation_photos".format(base_url=INAT_BASE_URL),
        access_token=access_token,
        data={"observation_photo[observation_id]": observation_id},
        files={"file": ensure_file_obj(photo)},
        user_agent=user_agent,
    )

    return response.json()


@document_request_params([docs._observation_id, docs._access_token])
def delete_observation(observation_id: int, access_token: str = None, user_agent: str = None):
    """
    Delete an observation.

    **API reference:** https://www.inaturalist.org/pages/api+reference#delete-observations-id

    Example:

        >>> token = get_access_token('...')
        >>> delete_observation(17932425, token)

    Returns:
        If successful, no response is returned from this endpoint

    Raises:
        :py:exc:`.ObservationNotFound` if the requested observation doesn't exist
        :py:exc:`requests.HTTPError` (403) if the observation belongs to another user
    """
    response = delete(
        url="{base_url}/observations/{id}.json".format(base_url=INAT_BASE_URL, id=observation_id),
        access_token=access_token,
        user_agent=user_agent,
        headers={"Content-type": "application/json"},
    )
    if response.status_code == 404:
        raise ObservationNotFound
    response.raise_for_status()



""" Some common functions for HTTP requests used by both the Node and REST API modules """
from logging import getLogger
from os import getenv
from typing import Dict, List, Union
from unittest.mock import Mock
from urllib.parse import urljoin

import requests

import pyinaturalist
from pyinaturalist.constants import WRITE_HTTP_METHODS
from pyinaturalist.request_params import preprocess_request_params, convert_list, validate_ids

# Mock response content to return in dry-run mode
MOCK_RESPONSE = Mock(spec=requests.Response)
MOCK_RESPONSE.json.return_value = {"results": [], "total_results": 0}

logger = getLogger(__name__)


# TODO: Copy function signature of request(), add `url`, and apply to these 4 wrapper functions
def delete(url: str, **kwargs) -> requests.Response:
    """ Wrapper around :py:func:`requests.delete` that supports dry-run mode """
    return request("DELETE", url, **kwargs)


def get(url: str, **kwargs) -> requests.Response:
    """ Wrapper around :py:func:`requests.get` that supports dry-run mode """
    return request("GET", url, **kwargs)


def post(url: str, **kwargs) -> requests.Response:
    """ Wrapper around :py:func:`requests.post` that supports dry-run mode """
    return request("POST", url, **kwargs)


def put(url: str, **kwargs) -> requests.Response:
    """ Wrapper around :py:func:`requests.put` that supports dry-run mode """
    return request("PUT", url, **kwargs)


def request(
    method: str,
    url: str,
    access_token: str = None,
    user_agent: str = None,
    ids: Union[str, List] = None,
    params: Dict = None,
    headers: Dict = None,
    **kwargs
) -> requests.Response:
    """Wrapper around :py:func:`requests.request` that supports dry-run mode and
    adds appropriate headers.

    Args:
        method: HTTP method
        url: Request URL
        access_token: access_token: the access token, as returned by :func:`get_access_token()`
        user_agent: a user-agent string that will be passed to iNaturalist
        ids: One or more integer IDs used as REST resource(s) to request
        params: Requests parameters
        headers: Request headers

    Returns:
        API response
    """
    # Set user agent and authentication headers, if specified
    headers = headers or {}
    headers["Accept"] = "application/json"
    headers["User-Agent"] = user_agent or pyinaturalist.user_agent
    if access_token:
        headers["Authorization"] = "Bearer %s" % access_token

    params = preprocess_request_params(params)

    # If one or more REST resources are requested, update the request URL accordignly
    if ids:
        url = url.rstrip("/") + "/" + validate_ids(ids)

    # Run either real request or mock request depending on settings
    if is_dry_run_enabled(method):
        logger.debug("Dry-run mode enabled; mocking request")
        log_request(method, url, params=params, headers=headers, **kwargs)
        return MOCK_RESPONSE
    else:
        return requests.request(method, url, params=params, headers=headers, **kwargs)


def is_dry_run_enabled(method: str) -> bool:
    """A wrapper to determine if dry-run (aka test mode) has been enabled via either
    a constant or an environment variable. Dry-run mode may be enabled for either write
    requests, or all requests.
    """
    dry_run_enabled = pyinaturalist.DRY_RUN_ENABLED or env_to_bool("DRY_RUN_ENABLED")
    if method in WRITE_HTTP_METHODS:
        return (
            dry_run_enabled or pyinaturalist.DRY_RUN_WRITE_ONLY or env_to_bool("DRY_RUN_WRITE_ONLY")
        )
    return dry_run_enabled


def env_to_bool(environment_variable: str) -> bool:
    """Translate an environment variable to a boolean value, accounting for minor
    variations (case, None vs. False, etc.)
    """
    env_value = getenv(environment_variable)
    return bool(env_value) and str(env_value).lower() not in ["false", "none"]


def log_request(*args, **kwargs):
    """ Log all relevant information about an HTTP request """
    kwargs_strs = ["{}={}".format(k, v) for k, v in kwargs.items()]
    logger.info("Request: {}".format(", ".join(list(args) + kwargs_strs)))



from datetime import date, datetime
from typing import Any, BinaryIO, Dict, List, Union

INAT_NODE_API_BASE_URL = "https://api.inaturalist.org/v1/"
INAT_BASE_URL = "https://www.inaturalist.org"

PER_PAGE_RESULTS = 30  # Number of records per page for paginated queries
THROTTLING_DELAY = 1  # In seconds, support <1 floats such as 0.1

# Toggle dry-run mode: this will run and log mock HTTP requests instead of real ones
DRY_RUN_ENABLED = False  # Mock all requests, including GET
DRY_RUN_WRITE_ONLY = False  # Only mock 'write' requests
WRITE_HTTP_METHODS = ["PATCH", "POST", "PUT", "DELETE"]

# Type aliases
Date = Union[date, datetime, str]
DateTime = Union[date, datetime, str]
FileOrPath = Union[BinaryIO, str]
IntOrStr = Union[int, str]
JsonResponse = Dict[str, Any]
ListResponse = List[Dict[str, Any]]
ObsFieldValues = Union[Dict, List[Dict]]
RequestParams = Dict[str, Any]
MultiInt = Union[int, List[int]]
MultiStr = Union[str, List[str]]
TemplateFunction = Any  # Cannot use Callable/Protocol, as they will not allow a mix of signatures



"""Helper functions for processing and validating request parameters"""
from datetime import date, datetime
from io import BytesIO
from logging import getLogger
from os.path import abspath, expanduser
from typing import Any, BinaryIO, Dict, Iterable, Optional
import warnings

from dateutil.parser import parse as parse_timestamp
from dateutil.tz import tzlocal
from pyinaturalist.constants import FileOrPath, RequestParams


# Basic observation attributes to include by default in geojson responses
DEFAULT_OBSERVATION_ATTRS = [
    "id",
    "photo_url",
    "positional_accuracy",
    "preferred_common_name",
    "quality_grade",
    "taxon_id",
    "taxon_name",
    "time_observed_at",
    "uri",
]

# All request parameters from both Node API and REST (Rails) API that accept date or datetime strings
DATETIME_PARAMS = [
    "created_after",
    "created_d1",
    "created_d2",
    "created_on",
    "d1",
    "d2",
    "newer_than",
    "observation_created_d1",
    "observation_created_d2",
    "observed_d1",
    "observed_d2",
    "observed_on",
    "older_than",
    "on",
    "since",
    "updated_since",  # TODO: test if this one behaves differently in Node API vs REST API
]

# Reponse formats supported by GET /observations endpoint
OBSERVATION_FORMATS = ["atom", "csv", "dwc", "json", "kml", "widget"]

# Creative Commons license codes
CC_LICENSES = ["CC-BY", "CC-BY-NC", "CC-BY-ND", "CC-BY-SA", "CC-BY-NC-ND", "CC-BY-NC-SA", "CC0"]

# IUCN Conservation status codes; for more info, see: https://www.iucnredlist.org
CONSERVATION_STATUSES = ["LC", "NT", "VU", "EN", "CR", "EW", "EX"]

# Taxon ID and name of main taxa "categories" that can be filtered on
ICONIC_TAXA = {
    0: "Unknown",
    1: "Animalia",
    3: "Aves",
    20978: "Amphibia",
    26036: "Reptilia",
    40151: "Mammalia",
    47178: "Actinopterygii",
    47115: "Mollusca",
    47119: "Arachnida",
    47158: "Insecta",
    47126: "Plantae",
    47170: "Fungi",
    48222: "Chromista",
    47686: "Protozoa",
}

# Taxonomic ranks that can be filtered on
RANKS = [
    "form",
    "variety",
    "subspecies",
    "hybrid",
    "species",
    "genushybrid",
    "subgenus",
    "genus",
    "subtribe",
    "tribe",
    "supertribe",
    "subfamily",
    "family",
    "epifamily",
    "superfamily",
    "infraorder",
    "suborder",
    "order",
    "superorder",
    "infraclass",
    "subclass",
    "class",
    "superclass",
    "subphylum",
    "phylum",
    "kingdom",
]

# Endpoint-specific options for multiple choice parameters
NODE_OBS_ORDER_BY_PROPERTIES = ["created_at", "id", "observed_on", "species_guess", "votes"]
REST_OBS_ORDER_BY_PROPERTIES = ["date_added", "observed_on"]
PROJECT_ORDER_BY_PROPERTIES = ["created", "distance", "featured", "recent_posts", "updated"]

# Options for multiple choice parameters (non-endpoint-specific)
COMMUNITY_ID_STATUSES = ["most_agree", "most_disagree", "some_agree"]
EXTRA_PROPERTIES = ["fields", "identifications", "projects"]
GEOPRIVACY_LEVELS = ["obscured", "obscured_private", "open", "private"]
HAS_PROPERTIES = ["photo", "geo"]
ORDER_DIRECTIONS = ["asc", "desc"]
PROJECT_TYPES = ["collection", "umbrella"]
QUALITY_GRADES = ["casual", "needs_id", "research"]
SEARCH_PROPERTIES = ["names", "tags", "description", "place"]


# Multiple-choice request parameters, with keys mapped to their possible choices (non-endpoint-specific)
MULTIPLE_CHOICE_PARAMS = {
    "csi": CONSERVATION_STATUSES,
    "extra": EXTRA_PROPERTIES,
    "geoprivacy": GEOPRIVACY_LEVELS,
    "has": HAS_PROPERTIES,
    "hrank": RANKS,
    "iconic_taxa": list(ICONIC_TAXA.values()),
    "identifications": COMMUNITY_ID_STATUSES,
    "license": CC_LICENSES,
    "lrank": RANKS,
    "max_rank": RANKS,
    "min_rank": RANKS,
    "order": ORDER_DIRECTIONS,
    "photo_license": CC_LICENSES,
    "quality_grade": QUALITY_GRADES,
    "rank": RANKS,
    "search_on": SEARCH_PROPERTIES,
    "sound_license": CC_LICENSES,
    "taxon_geoprivacy": GEOPRIVACY_LEVELS,
    "type": PROJECT_TYPES,
}

MULTIPLE_CHOICE_ERROR_MSG = (
    "Parameter '{}' must have one of the following values: {}\n\tValue provided: {}"
)

logger = getLogger(__name__)


def preprocess_request_params(params: Optional[Dict[str, Any]]) -> Dict[str, Any]:
    """ Perform type conversions, sanity checks, etc. on request parameters """
    if not params:
        return {}

    validate_multiple_choice_params(params)
    params = convert_bool_params(params)
    params = convert_datetime_params(params)
    params = convert_list_params(params)
    params = strip_empty_params(params)
    return params


# TODO: Remove in 0.12
def check_deprecated_params(params=None, **kwargs) -> Dict[str, Any]:
    """Check for usage of request parameters that are deprecated but still supported for
    backwards-compatibility
    """
    if params:
        warn("The 'params' argument is deprecated; please use keyword arguments instead")
        kwargs.update(params)
    if kwargs.get("search_query"):
        warn("The 'search_query' argument is deprecated; use 'q' instead")
        kwargs["q"] = kwargs.pop("search_query")

    return kwargs


def convert_bool_params(params: RequestParams) -> RequestParams:
    """ Convert any boolean request parameters to javascript-style boolean strings """
    for k, v in params.items():
        if isinstance(v, bool):
            params[k] = str(v).lower()
    return params


def convert_datetime_params(params: RequestParams) -> RequestParams:
    """Convert any dates, datetimes, or timestamps in other formats into ISO 8601 strings.

    API behavior note: params that take date but not time info will accept a full timestamp and
    just ignore the time, so it's safe to parse both date and datetime strings into timestamps

    Raises:
        :py:exc:`dateutil.parser._parser.ParserError` if a date/datetime format is invalid
    """
    for k, v in params.items():
        if isinstance(v, datetime) or isinstance(v, date):
            params[k] = _isoformat(v)
        if v is not None and k in DATETIME_PARAMS:
            params[k] = _isoformat(parse_timestamp(v))

    return params


def ensure_file_obj(photo: FileOrPath) -> BinaryIO:
    """Given a file objects or path, read it into a file-like object if it's a path"""
    if isinstance(photo, str):
        file_path = abspath(expanduser(photo))
        logger.info("Reading from file: {}".format(file_path))
        with open(file_path, "rb") as f:
            return BytesIO(f.read())
    return photo


def ensure_file_objs(photos: Iterable[FileOrPath]) -> Iterable[BinaryIO]:
    """Given one or more file objects and/or paths, read any paths into a file-like object"""
    return [ensure_file_obj(photo) for photo in ensure_list(photos)]


def ensure_list(values: Any):
    """If the value is a string or comma-separated list of values, convert it into a list"""
    if not values:
        values = []
    elif isinstance(values, str) and "," in values:
        values = values.split(",")
    elif not isinstance(values, list):
        values = [values]
    return values


def convert_list(obj: Any) -> str:
    """Convert list parameters into an API-compatible (comma-delimited) string"""
    if not obj:
        return ""
    if isinstance(obj, list):
        return ",".join(map(str, obj))
    return str(obj)


def convert_list_params(params: RequestParams) -> RequestParams:
    """Convert any list parameters into an API-compatible (comma-delimited) string.
    Will be url-encoded by requests. For example: `['k1', 'k2', 'k3'] -> k1%2Ck2%2Ck3`
    """
    return {k: convert_list(v) for k, v in params.items()}


def convert_observation_fields(params: RequestParams) -> RequestParams:
    """Translate simplified format of observation field values into API-compatible format"""
    if "observation_fields" in params:
        params["observation_field_values_attributes"] = params.pop("observation_fields")
    obs_fields = params.get("observation_field_values_attributes")
    if isinstance(obs_fields, dict):
        params["observation_field_values_attributes"] = [
            {"observation_field_id": k, "value": v} for k, v in obs_fields.items()
        ]
    return params


def strip_empty_params(params: RequestParams) -> RequestParams:
    """Remove any request parameters with empty or ``None`` values."""
    return {k: v for k, v in params.items() if v or v is False}


def is_int(value: Any) -> bool:
    """ Determine if a value is a valid integer """
    try:
        int(value)
        return True
    except (TypeError, ValueError):
        return False


def is_int_list(values: Any) -> bool:
    """Determine if a value contains one or more valid integers"""
    return all([is_int(v) for v in ensure_list(values)])


def validate_ids(ids: Any) -> str:
    """Ensure ID(s) are all valid integers, and convert to a comma-delimited string if multiple

    Raises:
        :py:exc:`ValueError` if any values are not valid integers
    """
    if not is_int_list(ids):
        raise ValueError("Invalid ID(s): {}; must specify integers only".format(ids))
    return convert_list([int(id) for id in ensure_list(ids)])


def is_valid_multiple_choice_option(value: Any, choices: Iterable) -> bool:
    """Determine if a multiple-choice request parameter contains valid value(s)."""
    if not value:
        return True
    if not isinstance(value, list):
        value = [value]
    return all([v in choices for v in value])


def validate_multiple_choice_param(params: RequestParams, key: str, choices: Iterable):
    """Verify if a multiple-choice request parameter contains valid value(s);
    if not, raise an error. **Used for endpoint-specific params.**

    Raises:
        :py:exc:`ValueError`
    """
    if key in params and not is_valid_multiple_choice_option(params[key], choices):
        raise ValueError(MULTIPLE_CHOICE_ERROR_MSG.format(key, choices, params[key]))


def validate_multiple_choice_params(params: RequestParams):
    """Verify that multiple-choice request parameters contain a valid value.

    **Note:** This does not check endpoint-specific params, i.e., those that have the same name
    but different values across different endpoints.

    Raises:
        :py:exc:`ValueError`;
            Error message will contain info on all validation errors, if there are multiple
    """
    # Collect info on any validation errors
    errors = []
    for key, choices in MULTIPLE_CHOICE_PARAMS.items():
        if key in params and not is_valid_multiple_choice_option(params[key], choices):
            errors.append(MULTIPLE_CHOICE_ERROR_MSG.format(key, choices, params[key]))

    # Combine all messages (if multiple) into one error message
    if errors:
        raise ValueError("\n".join(errors))


def _isoformat(d):
    """Return a date or datetime in ISO format.
    If it's a datetime and doesn't already have tzinfo, set it to the system's local timezone.
    """
    if isinstance(d, datetime) and not d.tzinfo:
        d = d.replace(tzinfo=tzlocal())
    return d.isoformat()


def translate_rank_range(params: RequestParams) -> RequestParams:
    """If min and/or max rank is specified in params, translate into a list of ranks"""

    def _get_rank_index(rank: str) -> int:
        if rank not in RANKS:
            raise ValueError("Invalid rank")
        return RANKS.index(rank)

    min_rank, max_rank = params.pop("min_rank", None), params.pop("max_rank", None)
    if min_rank or max_rank:
        # Use indices in RANKS list to determine range of ranks to include
        min_rank_index = _get_rank_index(min_rank) if min_rank else 0
        max_rank_index = _get_rank_index(max_rank) + 1 if max_rank else len(RANKS)
        params["rank"] = RANKS[min_rank_index:max_rank_index]
    return params


def warn(msg):
    warnings.warn(DeprecationWarning(msg))



import pytest
from datetime import date, datetime
from dateutil.tz import gettz
from io import BytesIO
from tempfile import NamedTemporaryFile
from unittest.mock import patch

from pyinaturalist.request_params import (
    preprocess_request_params,
    convert_bool_params,
    convert_datetime_params,
    convert_list_params,
    convert_observation_fields,
    ensure_file_obj,
    strip_empty_params,
    validate_ids,
    validate_multiple_choice_param,
    validate_multiple_choice_params,
)
import pyinaturalist.rest_api
import pyinaturalist.node_api
from test.conftest import get_module_http_functions, get_mock_args_for_signature


TEST_PARAMS = {
    "is_active": False,
    "only_id": "true",
    "preferred_place_id": [1, 2],
    "rank": ["phylum", "class"],
    "q": "",
    "locale": None,
    "parent_id": 1,
    "observation_fields": {1: "value"},
}


def test_convert_bool_params():
    params = convert_bool_params(TEST_PARAMS)
    assert params["is_active"] == "false"
    assert params["only_id"] == "true"


# Test some recognized date(time) formats, with and without TZ info, in date and non-date params
@pytest.mark.parametrize(
    "param, value, expected",
    [
        ("created_d1", "19951231T235959", "1995-12-31T23:59:59-08:00"),
        ("created_d2", "2008-08-08 08:08:08Z", "2008-08-08T08:08:08+00:00"),
        ("created_on", "2010-10-10 10:10:10-05:00", "2010-10-10T10:10:10-05:00"),
        ("created_on", "Jan 1 2000", "2000-01-01T00:00:00-08:00"),
        ("d1", "19970716", "1997-07-16T00:00:00-07:00"),
        ("q", date(1954, 2, 5), "1954-02-05"),
        ("q", datetime(1954, 2, 5), "1954-02-05T00:00:00-08:00"),
        ("q", "not a datetime", "not a datetime"),
    ],
)
@patch("pyinaturalist.request_params.tzlocal", return_value=gettz("US/Pacific"))
def test_convert_datetime_params(tzlocal, param, value, expected):
    converted = convert_datetime_params({param: value})
    assert converted[param] == expected


# Test both int and string lists
def test_convert_list_params():
    params = convert_list_params(TEST_PARAMS)
    assert params["preferred_place_id"] == "1,2"
    assert params["rank"] == "phylum,class"


def test_convert_observation_fields():
    params = convert_observation_fields(TEST_PARAMS)
    assert params["observation_field_values_attributes"] == [
        {"observation_field_id": 1, "value": "value"}
    ]


def test_ensure_file_obj__file_path():
    with NamedTemporaryFile() as temp:
        temp.write(b"test content")
        temp.seek(0)

        file_obj = ensure_file_obj(temp.name)
        assert file_obj.read() == b"test content"


def test_ensure_file_obj__file_obj():
    file_obj = BytesIO(b"test content")
    assert ensure_file_obj(file_obj) == file_obj


def test_strip_empty_params():
    params = strip_empty_params(TEST_PARAMS)
    assert len(params) == 6
    assert "q" not in params and "locale" not in params
    assert "is_active" in params and "only_id" in params


@pytest.mark.parametrize(
    "value, expected",
    [
        ("1", "1"),
        (1, "1"),
        ("1,2,3", "1,2,3"),
        ([1, 2, 3], "1,2,3"),
        ([1e5, 2e5], "100000,200000"),
    ],
)
def test_validate_ids(value, expected):
    assert validate_ids(value) == expected


@pytest.mark.parametrize("value", ["NaN", "", None, []])
def test_validate_ids__invalid(value):
    with pytest.raises(ValueError):
        validate_ids("not a number")


# This is just here so that tests will fail if one of the conversion steps is removed
@patch("pyinaturalist.request_params.convert_bool_params")
@patch("pyinaturalist.request_params.convert_datetime_params")
@patch("pyinaturalist.request_params.convert_list_params")
@patch("pyinaturalist.request_params.strip_empty_params")
def test_preprocess_request_params(mock_bool, mock_datetime, mock_list, mock_strip):
    preprocess_request_params({"id": 1})
    assert all([mock_bool.called, mock_datetime.called, mock_list.called, mock_strip.called])


# The following tests ensure that all API requests call preprocess_request_params() at some point
# Almost all logic except the request is mocked out so this can generically apply to all API functions
# Using parametrization here so that on failure, pytest will show the specific function that failed
@pytest.mark.parametrize(
    "http_function", get_module_http_functions(pyinaturalist.node_api).values()
)
@patch("pyinaturalist.request_params.translate_rank_range")
@patch("pyinaturalist.response_format.as_geojson_feature")
@patch("pyinaturalist.node_api.convert_location_to_float")
@patch("pyinaturalist.node_api._convert_all_locations_to_float")
@patch("pyinaturalist.api_requests.preprocess_request_params")
@patch("pyinaturalist.api_requests.requests.request")
def test_all_node_requests_use_param_conversion(
    request,
    preprocess_request_params,
    convert_all_locations_to_float,
    convert_location_to_float,
    as_geojson,
    translate_rank_range,
    http_function,
):
    request().json.return_value = {"total_results": 1, "results": [{}]}
    mock_args = get_mock_args_for_signature(http_function)
    http_function(*mock_args)
    assert preprocess_request_params.call_count == 1


@pytest.mark.parametrize(
    "http_function", get_module_http_functions(pyinaturalist.rest_api).values()
)
@patch("pyinaturalist.rest_api.convert_lat_long_to_float")
@patch("pyinaturalist.rest_api.sleep")
@patch("pyinaturalist.api_requests.preprocess_request_params")
@patch("pyinaturalist.api_requests.requests.request")
def test_all_rest_requests_use_param_conversion(
    request, preprocess_request_params, sleep, convert_lat_long_to_float, http_function
):
    # Handle the one API response that returns a list instead of a dict
    if http_function == pyinaturalist.rest_api.get_all_observation_fields:
        request().json.return_value = []
    else:
        request().json.return_value = {
            "total_results": 1,
            "access_token": "",
            "results": [],
        }

    mock_args = get_mock_args_for_signature(http_function)
    http_function(*mock_args)
    assert preprocess_request_params.call_count == 1


def test_validate_multiple_choice_param():
    params = {
        "param1": "valid_str",
        "param2": "invalid_str",
    }
    choices = ["str1", "str2", "valid_str"]

    validate_multiple_choice_param(params, "param1", choices)
    assert True
    with pytest.raises(ValueError):
        validate_multiple_choice_param(params, "param2", choices)


@pytest.mark.parametrize(
    "params",
    [
        {"csi": "LC"},
        {"csi": ["EW", "EX"]},
        {"geoprivacy": "open"},
        {"iconic_taxa": "Animalia"},
        {"identifications": "most_agree"},
        {"license": "CC-BY-NC"},
        {"rank": "order"},
        {"quality_grade": "research"},
        {"search_on": "tags"},
        {"geoprivacy": ["open", "obscured"]},
        {"geoprivacy": "open", "iconic_taxa": "Animalia", "license": "CC-BY-NC"},
    ],
)
def test_validate_multiple_choice_params(params):
    # If valid, there is no return value, but an exception should not be raised
    validate_multiple_choice_params(params)
    # If invalid, an exception should be raised
    with pytest.raises(ValueError):
        validate_multiple_choice_params({k: "Invalid_value" for k in params})
    # A valid + invalid value should also raise an exception
    with pytest.raises(ValueError):
        validate_multiple_choice_params({k: [v, "Invalid_value"] for k, v in params.items()})



"""
Shared unit test-related utilities.
Pytest will also automatically pick up any fixtures defined here.
"""
import json
import logging
import re
from inspect import getmembers, isfunction, signature, Parameter
from os.path import abspath, dirname, join
from unittest.mock import MagicMock

HTTP_FUNC_PATTERN = re.compile(r"(get|put|post|delete)_.+")
SAMPLE_DATA_DIR = abspath(join(dirname(__file__), "sample_data"))
# Enable logging for urllib and other external loggers
logging.basicConfig(level="INFO")


def get_module_functions(module):
    """ Get all functions belonging to a module (excluding imports) """
    return {
        name: member
        for name, member in getmembers(module)
        if isfunction(member) and member.__module__ == module.__name__
    }


def get_module_http_functions(module):
    """ Get all functions belonging to a module and prefixed with an HTTP method """
    return {
        name: func
        for name, func in get_module_functions(module).items()
        if HTTP_FUNC_PATTERN.match(name.lower())
    }


def get_mock_args_for_signature(func):
    """Automagically get a list of mock objects corresponding to the required arguments
    in a function's signature. Using ``inspect.Signature``, 'Required' is defined by:
    1. positional and 2. no default
    """
    required_args = [
        p
        for p in signature(func).parameters.values()
        if p.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
        and p.default is Parameter.empty
    ]
    return [MagicMock()] * len(required_args)


def _sample_data_path(filename):
    return join(SAMPLE_DATA_DIR, filename)


def load_sample_data(filename):
    with open(_sample_data_path(filename), encoding="utf-8") as fh:
        if filename.endswith("json"):
            return json.load(fh)
        else:
            return fh.read()



#!/usr/bin/env python
from itertools import chain
from setuptools import setup, find_packages
from pyinaturalist import __version__

# These package categories allow tox and build environments to install only what they need
extras_require = {
    # Packages used for CI jobs
    "build": ["coveralls", "tox-travis"],
    # Packages used for documentation builds
    "docs": [
        "m2r2",
        "Sphinx~=3.2.1",
        "sphinx-autodoc-typehints",
        "sphinx-automodapi",
        "sphinx-rtd-theme",
        "sphinxcontrib-apidoc",
    ],
    # Packages used for testing both locally and in CI jobs
    "test": [
        "black==20.8b1",
        "flake8",
        "mypy",
        "pytest>=5.0",
        "pytest-cov",
        "requests-mock>=1.7",
        "tox>=3.15",
    ],
}
# All development/testing packages combined
extras_require["dev"] = list(chain.from_iterable(extras_require.values()))


setup(
    name="pyinaturalist",
    version=__version__,
    author="Nicolas Noé",
    author_email="nicolas@niconoe.eu",
    url="https://github.com/niconoe/pyinaturalist",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["python-dateutil>=2.0", "python-forge", "requests>=2.24.0"],
    extras_require=extras_require,
    zip_safe=False,
)



from pyinaturalist.forge_utils import document_request_params, copy_docstrings, copy_signatures


def test_document_request_params():
    pass


def func_template():
    """
    arg_1: First test arg
    arg_2: Second test arg
    """


def test_copy_docstrings__with_return_desc():
    def func():
        """Test function.

        Returns:
            None
        """

    expected_docstring = """
Test function.

Args:
    arg_1: First test arg
    arg_2: Second test arg

Returns:
    None
    """.strip()

    modified_func = copy_docstrings(func, [func_template])
    assert modified_func.__doc__ == expected_docstring


def test_copy_docstrings__without_return_desc():
    def func():
        """ Test function. """

    expected_docstring = """
Test function.

Args:
    arg_1: First test arg
    arg_2: Second test arg
    """.strip()

    modified_func = copy_docstrings(func, [func_template])
    assert modified_func.__doc__ == expected_docstring


def test_copy_signatures():
    pass



import pytest
from unittest.mock import patch

import pyinaturalist
from pyinaturalist.api_requests import MOCK_RESPONSE, delete, get, post, put, request


# Just test that the wrapper methods call requests.request with the appropriate HTTP method
@pytest.mark.parametrize(
    "function, http_method",
    [(delete, "DELETE"), (get, "GET"), (post, "POST"), (put, "PUT")],
)
@patch("pyinaturalist.api_requests.request")
def test_http_methods(mock_request, function, http_method):
    function("https://url", param="value")
    mock_request.assert_called_with(http_method, "https://url", param="value")


# Test that the requests() wrapper passes along expected headers; just tests kwargs, not mock response
@pytest.mark.parametrize(
    "input_kwargs, expected_headers",
    [
        ({}, {"Accept": "application/json", "User-Agent": pyinaturalist.user_agent}),
        (
            {"user_agent": "CustomUA"},
            {"Accept": "application/json", "User-Agent": "CustomUA"},
        ),
        (
            {"access_token": "token"},
            {
                "Accept": "application/json",
                "User-Agent": pyinaturalist.user_agent,
                "Authorization": "Bearer token",
            },
        ),
    ],
)
@patch("pyinaturalist.api_requests.requests.request")
def test_request_headers(mock_request, input_kwargs, expected_headers):
    request("GET", "https://url", **input_kwargs)
    request_kwargs = mock_request.call_args[1]
    assert request_kwargs["headers"] == expected_headers


# Test relevant combinations of dry-run settings and HTTP methods
@pytest.mark.parametrize(
    "enabled_const, enabled_env, write_only_const, write_only_env, method, expected_real_request",
    [
        # DRY_RUN_ENABLED constant or envar should mock GETs, but not DRY_RUN_WRITE_ONLY
        (False, False, False, False, "GET", True),
        (False, False, True, True, "GET", True),
        (False, False, False, False, "HEAD", True),
        (True, False, False, False, "GET", False),
        (False, True, False, False, "GET", False),
        # Either DRY_RUN_ENABLED or DRY_RUN_WRITE_ONLY should mock POST requests
        (False, False, False, False, "POST", True),
        (True, False, False, False, "POST", False),
        (False, True, False, False, "POST", False),
        (False, False, True, False, "POST", False),
        (False, False, False, True, "POST", False),
        (False, False, True, False, "POST", False),
        # Same for the other write methods
        (False, False, False, False, "PUT", True),
        (False, False, False, False, "DELETE", True),
        (False, False, False, True, "PUT", False),
        (False, False, False, True, "DELETE", False),
        # Truthy environment variable strings should be respected
        (False, "true", False, False, "GET", False),
        (False, "True", False, "False", "PUT", False),
        (False, False, False, "True", "DELETE", False),
        # As well as "falsy" environment variable strings
        (False, "false", False, False, "GET", True),
        (False, "none", False, "False", "POST", True),
        (False, False, False, "None", "DELETE", True),
    ],
)
@patch("pyinaturalist.api_requests.getenv")
@patch("pyinaturalist.api_requests.requests")
def test_request_dry_run(
    mock_requests,
    mock_getenv,
    enabled_const,
    enabled_env,
    write_only_const,
    write_only_env,
    method,
    expected_real_request,
):
    # Mock any environment variables specified
    env_vars = {}
    if enabled_env is not None:
        env_vars["DRY_RUN_ENABLED"] = enabled_env
    if write_only_env is not None:
        env_vars["DRY_RUN_WRITE_ONLY"] = write_only_env
    mock_getenv.side_effect = env_vars.__getitem__

    # Mock constants and run request
    with patch("pyinaturalist.api_requests.pyinaturalist") as settings:
        settings.DRY_RUN_ENABLED = enabled_const
        settings.DRY_RUN_WRITE_ONLY = write_only_const
        response = request(method, "http://url")

    # Verify that the request was or wasn't mocked based on settings
    if expected_real_request:
        assert mock_requests.request.call_count == 1
        assert response == mock_requests.request()
    else:
        assert response == MOCK_RESPONSE
        assert mock_requests.request.call_count == 0


# In addition to the test cases above, ensure that the request/response isn't altered with dry-run disabled
def test_request_dry_run_disabled(requests_mock):
    real_response = {"results": ["I'm a real response object!"]}
    requests_mock.get(
        "http://url", json={"results": ["I'm a real response object!"]}, status_code=200
    )

    assert request("GET", "http://url").json() == real_response



import pytest
from urllib.parse import urlencode, urljoin
from unittest.mock import patch

import pyinaturalist
from pyinaturalist.constants import INAT_NODE_API_BASE_URL, PER_PAGE_RESULTS
from pyinaturalist.node_api import (
    get_observation,
    get_observations,
    get_all_observations,
    get_observation_species_counts,
    get_all_observation_species_counts,
    get_geojson_observations,
    get_places_by_id,
    get_places_nearby,
    get_places_autocomplete,
    get_projects,
    get_projects_by_id,
    get_taxa,
    get_taxa_by_id,
    get_taxa_autocomplete,
)
from pyinaturalist.exceptions import ObservationNotFound
from pyinaturalist.rest_api import get_access_token
from test.conftest import load_sample_data


# Observations
# --------------------


def test_get_observation(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "observations"),
        json=load_sample_data("get_observation.json"),
        status_code=200,
    )

    observation = get_observation(16227955)
    assert observation["quality_grade"] == "research"
    assert observation["id"] == 16227955
    assert observation["user"]["login"] == "niconoe"
    assert len(observation["photos"]) == 2


def test_get_observations(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "observations"),
        json=load_sample_data("get_observations_node_page1.json"),
        status_code=200,
    )

    observations = get_observations(
        taxon_name="Danaus plexippus",
        created_on="2020-08-27",
        photos=True,
        geo=True,
        geoprivacy="open",
        place_id=7953,
    )
    first_result = observations["results"][0]

    assert observations["total_results"] == 2
    assert len(observations["results"]) == 1
    assert first_result["taxon_geoprivacy"] == "open"
    assert first_result["observed_on"] == "2020-08-27"
    assert first_result["taxon"]["id"] == 48662
    assert len(first_result["place_ids"]) == 13


@patch("pyinaturalist.node_api.sleep")
def test_get_all_observations(sleep, requests_mock):
    # Make response appear as though there are more pages
    page_1 = load_sample_data("get_observations_node_page1.json")
    page_1["total_results"] = PER_PAGE_RESULTS + 1
    page_2 = load_sample_data("get_observations_node_page2.json")

    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "observations"),
        [
            {"json": page_1, "status_code": 200},
            {"json": page_2, "status_code": 200},
        ],
    )

    observations = get_all_observations(id=[57754375, 57707611])

    assert type(observations) is list
    assert len(observations) == 2


def test_get_geojson_observations(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "observations"),
        json=load_sample_data("get_observation.json"),
        status_code=200,
    )

    geojson = get_geojson_observations(id=16227955)
    feature = geojson["features"][0]
    assert feature["geometry"]["coordinates"] == [4.360086, 50.646894]
    assert feature["properties"]["id"] == 16227955
    assert feature["properties"]["taxon_id"] == 493595


def test_get_geojson_observations__custom_properties(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "observations"),
        json=load_sample_data("get_observation.json"),
        status_code=200,
    )

    properties = ["taxon_name", "taxon_rank"]
    geojson = get_geojson_observations(id=16227955, properties=properties)
    feature = geojson["features"][0]
    assert feature["properties"]["taxon_name"] == "Lixus bardanae"
    assert feature["properties"]["taxon_rank"] == "species"
    assert "id" not in feature["properties"] and "taxon_id" not in feature["properties"]


def test_get_non_existent_observation(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "observations"),
        json=load_sample_data("get_nonexistent_observation.json"),
        status_code=200,
    )
    with pytest.raises(ObservationNotFound):
        get_observation(99999999)


def test_get_observation_species_counts(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "observations/species_counts"),
        json=load_sample_data("get_observation_species_counts.json"),
        status_code=200,
    )
    response = get_observation_species_counts(user_login="my_username", quality_grade="research")
    first_result = response["results"][0]

    assert first_result["count"] == 31
    assert first_result["taxon"]["id"] == 48484
    assert first_result["taxon"]["name"] == "Harmonia axyridis"


def test_get_all_observation_species_counts(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "observations/species_counts"),
        [
            {
                "json": load_sample_data("get_all_observation_species_counts_page1.json"),
                "status_code": 200,
            },
            {
                "json": load_sample_data("get_all_observation_species_counts_page2.json"),
                "status_code": 200,
            },
        ],
    )
    response = get_all_observation_species_counts(
        user_login="my_username", quality_grade="research"
    )
    first_result = response[0]
    last_result = response[-1]

    assert first_result["count"] == 19
    assert first_result["taxon"]["id"] == 27805
    assert first_result["taxon"]["name"] == "Notophthalmus viridescens"
    assert last_result["count"] == 1
    assert last_result["taxon"]["id"] == 39556
    assert last_result["taxon"]["name"] == "Apalone spinifera"


def test_get_observation_species_counts__invalid_multiple_choice_params():
    with pytest.raises(ValueError):
        get_observation_species_counts(quality_grade="None", iconic_taxa="slime molds")


# Places
# --------------------


def test_get_places_by_id(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "places/93735,89191"),
        json=load_sample_data("get_places_by_id.json"),
        status_code=200,
    )

    response = get_places_by_id([93735, 89191])
    result = response["results"][0]

    assert response["total_results"] == len(response["results"]) == 2
    assert result["id"] == 93735
    assert result["name"] == "Springbok"
    assert result["bbox_area"] == 0.000993854049
    assert result["location"] == [-29.665119, 17.88583]
    assert len(result["ancestor_place_ids"]) == 4


@pytest.mark.parametrize("place_id", ["asdf", [None], [1, "not a number"]])
def test_get_places_by_id__invalid_inputs(place_id):
    with pytest.raises(ValueError):
        get_places_by_id(place_id)


def test_get_places_nearby(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "places/nearby"),
        json=load_sample_data("get_places_nearby.json"),
        status_code=200,
    )

    response = get_places_nearby(nelat=150.0, nelng=-50.0, swlat=-149.999, swlng=-49.999)
    result = response["results"]["standard"][0]

    assert response["total_results"] == 20
    assert len(response["results"]["standard"]) + len(response["results"]["community"]) == 20

    assert result["id"] == 97394
    assert result["admin_level"] == -1
    assert result["name"] == "North America"
    assert result["bbox_area"] == 28171.40875125
    assert result["location"] == [56.7732555574, -179.68825]
    assert result["ancestor_place_ids"] is None


def test_get_places_autocomplete(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "places/autocomplete"),
        json=load_sample_data("get_places_autocomplete.json"),
        status_code=200,
    )

    response = get_places_autocomplete("springbok")
    result = response["results"][0]

    assert response["total_results"] == len(response["results"]) == 1
    assert result["id"] == 93735
    assert result["name"] == "Springbok"
    assert result["bbox_area"] == 0.000993854049
    assert result["location"] == [-29.665119, 17.88583]
    assert len(result["ancestor_place_ids"]) == 4


# Projects
# --------------------


def test_get_projects(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "projects"),
        json=load_sample_data("get_projects.json"),
        status_code=200,
    )

    response = get_projects(q="invasive", lat=49.27, lng=-123.08, radius=400, order_by="distance")
    first_result = response["results"][0]

    assert response["total_results"] == len(response["results"]) == 5
    assert first_result["id"] == 8291
    assert first_result["title"] == "PNW Invasive Plant EDDR"
    assert first_result["is_umbrella"] is False
    assert len(first_result["user_ids"]) == 33


def test_get_projects_by_id(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "projects/8348,6432"),
        json=load_sample_data("get_projects_by_id.json"),
        status_code=200,
    )
    response = get_projects_by_id([8348, 6432])
    first_result = response["results"][0]

    assert response["total_results"] == len(response["results"]) == 2
    assert first_result["id"] == 8348
    assert first_result["title"] == "Tucson High Native and Invasive Species Inventory"
    assert first_result["place_id"] == 96103
    assert first_result["location"] == [32.2264416406, -110.9617278383]


# Taxa
# --------------------

CLASS_AND_HIGHER = ["class", "superclass", "subphylum", "phylum", "kingdom"]
SPECIES_AND_LOWER = ["form", "variety", "subspecies", "hybrid", "species"]
CLASS_THOUGH_PHYLUM = ["class", "superclass", "subphylum", "phylum"]


def test_get_taxa(requests_mock):
    params = urlencode({"q": "vespi", "rank": "genus,subgenus,species"})
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "taxa?" + params),
        json=load_sample_data("get_taxa.json"),
        status_code=200,
    )

    response = get_taxa(q="vespi", rank=["genus", "subgenus", "species"])
    first_result = response["results"][0]

    assert len(response["results"]) == 30
    assert response["total_results"] == 35
    assert first_result["id"] == 70118
    assert first_result["name"] == "Nicrophorus vespilloides"
    assert first_result["rank"] == "species"
    assert first_result["is_active"] is True
    assert len(first_result["ancestor_ids"]) == 14


@pytest.mark.parametrize(
    "params, expected_ranks",
    [
        ({"rank": "genus"}, "genus"),
        ({"min_rank": "class"}, CLASS_AND_HIGHER),
        ({"max_rank": "species"}, SPECIES_AND_LOWER),
        ({"min_rank": "class", "max_rank": "phylum"}, CLASS_THOUGH_PHYLUM),
        ({"max_rank": "species", "rank": "override_me"}, SPECIES_AND_LOWER),
    ],
)
@patch("pyinaturalist.node_api.make_inaturalist_api_get_call")
def test_get_taxa_by_rank_range(
    mock_inaturalist_api_get_call,
    params,
    expected_ranks,
):
    # Make sure custom rank params result in the correct 'rank' param value
    get_taxa(**params)
    kwargs = mock_inaturalist_api_get_call.call_args[1]
    requested_rank = kwargs["params"]["rank"]
    assert requested_rank == expected_ranks


# This is just a spot test of a case in which boolean params should be converted
@patch("pyinaturalist.api_requests.requests.request")
def test_get_taxa_by_name_and_is_active(request):
    get_taxa(q="Lixus bardanae", is_active=False)
    request_kwargs = request.call_args[1]
    assert request_kwargs["params"] == {"q": "Lixus bardanae", "is_active": "false"}


def test_get_taxa_by_id(requests_mock):
    taxon_id = 70118
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "taxa/" + str(taxon_id)),
        json=load_sample_data("get_taxa_by_id.json"),
        status_code=200,
    )

    response = get_taxa_by_id(taxon_id)
    result = response["results"][0]
    assert response["total_results"] == len(response["results"]) == 1
    assert result["id"] == taxon_id
    assert result["name"] == "Nicrophorus vespilloides"
    assert result["rank"] == "species"
    assert result["is_active"] is True
    assert len(result["ancestors"]) == 12


@pytest.mark.parametrize("taxon_id", ["asdf", [None], [1, "not a number"]])
def test_get_taxa_by_id__invalid_inputs(taxon_id):
    with pytest.raises(ValueError):
        get_taxa_by_id(taxon_id)


def test_get_taxa_autocomplete(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "taxa/autocomplete"),
        json=load_sample_data("get_taxa_autocomplete.json"),
        status_code=200,
    )

    response = get_taxa_autocomplete(q="vespi")
    first_result = response["results"][0]

    assert len(response["results"]) == 10
    assert response["total_results"] == 47
    assert first_result["matched_term"] == "Vespidae"
    assert first_result["id"] == 52747
    assert first_result["name"] == "Vespidae"
    assert first_result["rank"] == "family"
    assert first_result["is_active"] is True
    assert len(first_result["ancestor_ids"]) == 11


# Test usage of format_taxon() with get_taxa_autocomplete()
def test_get_taxa_autocomplete_minified(requests_mock):
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "taxa/autocomplete"),
        json=load_sample_data("get_taxa_autocomplete.json"),
        status_code=200,
    )

    expected_results = [
        "   52747:       Family Vespidae (Hornets, Paper Wasps, Potter Wasps, and Allies)",
        "   84738:    Subfamily Vespinae (Hornets and Yellowjackets)",
        "  131878:      Species Nicrophorus vespillo (Vespillo Burying Beetle)",
        "  621585:      Species Vespicula trachinoides (Vespicula Waspfish)",
        "  495392:      Species Vespidae st1",
        "   70118:      Species Nicrophorus vespilloides (Lesser Vespillo Burying Beetle)",
        "   92786:        Genus Vespicula",
        "  646195:        Genus Vespiodes",
        "  621584:      Species Vespicula cypho",
        "  621586:      Species Vespicula zollingeri",
    ]

    response = get_taxa_autocomplete(q="vespi", minify=True)
    print(response["results"])
    assert response["results"] == expected_results


def test_user_agent(requests_mock):
    # TODO: test for all functions that access the inaturalist API?
    requests_mock.get(
        urljoin(INAT_NODE_API_BASE_URL, "observations"),
        json=load_sample_data("get_observation.json"),
        status_code=200,
    )
    accepted_json = {
        "access_token": "604e5df329b98eecd22bb0a84f88b68a075a023ac437f2317b02f3a9ba414a08",
        "token_type": "Bearer",
        "scope": "write",
        "created_at": 1539352135,
    }
    requests_mock.post(
        "https://www.inaturalist.org/oauth/token",
        json=accepted_json,
        status_code=200,
    )

    default_ua = "Pyinaturalist/{v}".format(v=pyinaturalist.__version__)

    # By default, we have a 'Pyinaturalist' user agent:
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == default_ua
    get_access_token("valid_username", "valid_password", "valid_app_id", "valid_app_secret")
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == default_ua

    # But if the user sets a custom one, it is indeed used:
    get_observation(16227955, user_agent="CustomUA")
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == "CustomUA"
    get_access_token(
        "valid_username",
        "valid_password",
        "valid_app_id",
        "valid_app_secret",
        user_agent="CustomUA",
    )
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == "CustomUA"

    # We can also set it globally:
    pyinaturalist.user_agent = "GlobalUA"
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == "GlobalUA"
    get_access_token("valid_username", "valid_password", "valid_app_id", "valid_app_secret")
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == "GlobalUA"

    # And it persists across requests:
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == "GlobalUA"
    get_access_token("valid_username", "valid_password", "valid_app_id", "valid_app_secret")
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == "GlobalUA"

    # But if we have a global and local one, the local has priority
    get_observation(16227955, user_agent="CustomUA 2")
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == "CustomUA 2"
    get_access_token(
        "valid_username",
        "valid_password",
        "valid_app_id",
        "valid_app_secret",
        user_agent="CustomUA 2",
    )
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == "CustomUA 2"

    # We can reset the global settings to the default:
    pyinaturalist.user_agent = pyinaturalist.DEFAULT_USER_AGENT
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == default_ua
    get_access_token("valid_username", "valid_password", "valid_app_id", "valid_app_secret")
    assert requests_mock._adapter.last_request._request.headers["User-Agent"] == default_ua



from datetime import datetime, timedelta
from io import BytesIO
from unittest.mock import patch

import pytest
from requests import HTTPError
from urllib.parse import urljoin

from pyinaturalist.constants import INAT_BASE_URL
from pyinaturalist.rest_api import (
    OBSERVATION_FORMATS,
    get_access_token,
    get_all_observation_fields,
    get_observations,
    get_observation_fields,
    put_observation_field_values,
    create_observation,
    update_observation,
    add_photo_to_observation,
    delete_observation,
)
from pyinaturalist.exceptions import AuthenticationError, ObservationNotFound
from test.conftest import load_sample_data

PAGE_1_JSON_RESPONSE = load_sample_data("get_observation_fields_page1.json")
PAGE_2_JSON_RESPONSE = load_sample_data("get_observation_fields_page2.json")


def get_observations_response(response_format):
    response_format = response_format.replace("widget", "js")
    return load_sample_data("get_observations.{}".format(response_format))


@pytest.mark.parametrize("response_format", OBSERVATION_FORMATS)
def test_get_observations(response_format, requests_mock):
    """ Test all supported observation data formats """
    response = get_observations_response(response_format)
    key = "json" if response_format == "json" else "text"

    requests_mock.get(
        urljoin(INAT_BASE_URL, "observations.{}".format(response_format)),
        status_code=200,
        **{key: response},
    )

    observations = get_observations(taxon_id=493595, response_format=response_format)

    # Ensure coordinate strings were converted to floats, for JSON format only
    if response_format == "json":
        response[0]["latitude"] = 50.646894
        response[0]["longitude"] = 4.360086
    assert observations == response


@pytest.mark.parametrize("response_format", ["geojson", "yaml"])
def test_get_observations__invalid_format(response_format):
    with pytest.raises(ValueError):
        get_observations(taxon_id=493595, response_format=response_format)


def test_get_observation_fields(requests_mock):
    """ get_observation_fields() work as expected (basic use)"""

    requests_mock.get(
        urljoin(INAT_BASE_URL, "observation_fields.json?q=sex&page=2"),
        json=PAGE_2_JSON_RESPONSE,
        status_code=200,
    )

    obs_fields = get_observation_fields(q="sex", page=2)
    assert obs_fields == PAGE_2_JSON_RESPONSE


def test_get_all_observation_fields(requests_mock):
    """get_all_observation_fields() is able to paginate, accepts a search query and return correct results"""

    requests_mock.get(
        urljoin(INAT_BASE_URL, "observation_fields.json?q=sex&page=1"),
        json=PAGE_1_JSON_RESPONSE,
        status_code=200,
    )

    requests_mock.get(
        urljoin(INAT_BASE_URL, "observation_fields.json?q=sex&page=2"),
        json=PAGE_2_JSON_RESPONSE,
        status_code=200,
    )

    page_3_json_response = []
    requests_mock.get(
        urljoin(INAT_BASE_URL, "observation_fields.json?q=sex&page=3"),
        json=page_3_json_response,
        status_code=200,
    )

    all_fields = get_all_observation_fields(q="sex")
    assert all_fields == PAGE_1_JSON_RESPONSE + PAGE_2_JSON_RESPONSE


def test_get_all_observation_fields_noparam(requests_mock):
    """get_all_observation_fields() can also be called without a search query without errors"""
    requests_mock.get(
        urljoin(INAT_BASE_URL, "observation_fields.json?page=1"),
        json=[],
        status_code=200,
    )

    get_all_observation_fields()


def test_put_observation_field_values(requests_mock):
    requests_mock.put(
        urljoin(INAT_BASE_URL, "observation_field_values/31"),
        json=load_sample_data("put_observation_field_value_result.json"),
        status_code=200,
    )

    r = put_observation_field_values(
        observation_id=18166477,
        observation_field_id=31,  # Animal behavior
        value="fouraging",
        access_token="valid token",
    )

    assert r["id"] == 31
    assert r["observation_field_id"] == 31
    assert r["observation_id"] == 18166477
    assert r["value"] == "fouraging"


def test_get_access_token_fail(requests_mock):
    """ If we provide incorrect credentials to get_access_token(), an AuthenticationError is raised"""

    rejection_json = {
        "error": "invalid_client",
        "error_description": "Client authentication failed due to "
        "unknown client, no client authentication "
        "included, or unsupported authentication "
        "method.",
    }
    requests_mock.post(
        urljoin(INAT_BASE_URL, "oauth/token"),
        json=rejection_json,
        status_code=401,
    )

    with pytest.raises(AuthenticationError):
        get_access_token("username", "password", "app_id", "app_secret")


def test_get_access_token(requests_mock):
    """ Test a successful call to get_access_token() """

    accepted_json = {
        "access_token": "604e5df329b98eecd22bb0a84f88b68a075a023ac437f2317b02f3a9ba414a08",
        "token_type": "Bearer",
        "scope": "write",
        "created_at": 1539352135,
    }
    requests_mock.post(
        urljoin(INAT_BASE_URL, "oauth/token"),
        json=accepted_json,
        status_code=200,
    )

    token = get_access_token("valid_username", "valid_password", "valid_app_id", "valid_app_secret")

    assert token == "604e5df329b98eecd22bb0a84f88b68a075a023ac437f2317b02f3a9ba414a08"


def test_update_observation(requests_mock):
    requests_mock.put(
        urljoin(INAT_BASE_URL, "observations/17932425.json"),
        json=load_sample_data("update_observation_result.json"),
        status_code=200,
    )

    params = {
        "ignore_photos": 1,
        "description": "updated description v2 !",
    }
    r = update_observation(17932425, access_token="valid token", **params)

    # If all goes well we got a single element representing the updated observation, enclosed in a list.
    assert len(r) == 1
    assert r[0]["id"] == 17932425
    assert r[0]["description"] == "updated description v2 !"


@patch("pyinaturalist.rest_api.ensure_file_objs")
@patch("pyinaturalist.rest_api.put")
def test_update_observation__local_photo(put, ensure_file_objs):
    update_observation(1234, access_token="token", local_photos="photo.jpg")

    # Make sure local_photos is replaced with the output of ensure_file_objs
    called_params = put.call_args[1]["json"]["observation"]
    assert called_params["local_photos"] == ensure_file_objs.return_value


def test_update_nonexistent_observation(requests_mock):
    """When we try to update a non-existent observation, iNat returns an error 410 with "obs does not longer exists". """
    requests_mock.put(
        urljoin(INAT_BASE_URL, "observations/999999999.json"),
        json={"error": "Cette observation n’existe plus."},
        status_code=410,
    )

    params = {
        "ignore_photos": 1,
        "description": "updated description v2 !",
    }

    with pytest.raises(HTTPError) as excinfo:
        update_observation(999999999, access_token="valid token", **params)
    assert excinfo.value.response.status_code == 410
    assert excinfo.value.response.json() == {"error": "Cette observation n’existe plus."}


def test_update_observation_not_mine(requests_mock):
    """When we try to update the obs of another user, iNat returns an error 410 with "obs does not longer exists"."""
    requests_mock.put(
        urljoin(INAT_BASE_URL, "observations/16227955.json"),
        json={"error": "Cette observation n’existe plus."},
        status_code=410,
    )

    params = {
        "ignore_photos": 1,
        "description": "updated description v2 !",
    }

    with pytest.raises(HTTPError) as excinfo:
        update_observation(16227955, access_token="valid token for another user", **params)
    assert excinfo.value.response.status_code == 410
    assert excinfo.value.response.json() == {"error": "Cette observation n’existe plus."}


def test_create_observation(requests_mock):
    requests_mock.post(
        urljoin(INAT_BASE_URL, "observations.json"),
        json=load_sample_data("create_observation_result.json"),
        status_code=200,
    )

    r = create_observation(species_guess="Pieris rapae", access_token="valid token")
    assert len(r) == 1  # We added a single one
    assert r[0]["latitude"] is None
    assert r[0]["taxon_id"] == 55626  # Pieris Rapae @ iNaturalist


@patch("pyinaturalist.rest_api.ensure_file_objs")
@patch("pyinaturalist.rest_api.post")
def test_create_observation__local_photo(post, ensure_file_objs):
    create_observation(access_token="token", local_photos="photo.jpg")

    # Make sure local_photos is replaced with the output of ensure_file_objs
    called_params = post.call_args[1]["json"]["observation"]
    assert called_params["local_photos"] == ensure_file_objs.return_value


def test_create_observation_fail(requests_mock):
    params = {
        "species_guess": "Pieris rapae",
        # Some invalid data so the observation is rejected...
        "observed_on_string": (datetime.now() + timedelta(days=1)).isoformat(),
        "latitude": 200,
    }

    requests_mock.post(
        urljoin(INAT_BASE_URL, "observations.json"),
        json=load_sample_data("create_observation_fail.json"),
        status_code=422,
    )

    with pytest.raises(HTTPError) as excinfo:
        create_observation(access_token="valid token", **params)
    assert excinfo.value.response.status_code == 422
    assert "errors" in excinfo.value.response.json()  # iNat also give details about the errors


def test_add_photo_to_observation(requests_mock):
    requests_mock.post(
        urljoin(INAT_BASE_URL, "observation_photos"),
        json=load_sample_data("add_photo_to_observation.json"),
        status_code=200,
    )

    response = add_photo_to_observation(1234, BytesIO(), access_token="token")
    assert response["id"] == 1234
    assert response["created_at"] == "2020-09-24T21:06:16.964-05:00"
    assert response["photo"]["native_username"] == "username"


def test_delete_observation(requests_mock):
    requests_mock.delete(urljoin(INAT_BASE_URL, "observations/24774619.json"), status_code=200)
    response = delete_observation(observation_id=24774619, access_token="valid token")
    assert response is None


def test_delete_unexisting_observation(requests_mock):
    """ObservationNotFound is raised if the observation doesn't exists"""
    requests_mock.delete(urljoin(INAT_BASE_URL, "observations/24774619.json"), status_code=404)

    with pytest.raises(ObservationNotFound):
        delete_observation(observation_id=24774619, access_token="valid token")
